# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import pickle
from pathlib import Path

import numpy as np
import pandas as pd

from asreview import __version__ as asreview_version
from asreview.config import LABEL_NA
from asreview.data import ASReviewData
from asreview.state.paths import get_data_file_path
from asreview.state.paths import get_labeled_path
from asreview.state.paths import get_pool_path
from asreview.webapp.utils.project_path import get_project_path


class CacheDataError(Exception):
    pass


def _get_cache_data_path(project_id):
    project_path = get_project_path(project_id)
    fp_data = get_data_file_path(project_path)

    return get_data_file_path(project_path) \
        .with_suffix(fp_data.suffix + ".pickle")


def _read_data_from_cache(project_id, version_check=True):

    fp_data_pickle = _get_cache_data_path(project_id)

    try:
        # get the pickle data
        with open(fp_data_pickle, 'rb') as f_pickle_read:
            data_obj, data_obj_version = pickle.load(f_pickle_read)

        # validate data object
        if not isinstance(data_obj.df, pd.DataFrame):
            raise ValueError()

        # drop cache files generated by older versions
        if (not version_check) or (asreview_version == data_obj_version):
            return data_obj

    except FileNotFoundError:
        # file not available
        pass
    except Exception as err:
        # problem loading pickle file or outdated
        # remove the pickle file
        logging.error(f"Error reading cache file: {err}")
        try:
            os.remove(fp_data_pickle)
        except FileNotFoundError:
            pass

    raise CacheDataError()


def _write_data_to_cache(project_id, data_obj):

    fp_data_pickle = _get_cache_data_path(project_id)

    logging.info("Store a copy of the data in a pickle file.")
    with open(fp_data_pickle, 'wb') as f_pickle:
        pickle.dump((data_obj, asreview_version), f_pickle)


def read_data(project_id, use_cache=True, save_cache=True):
    """Get ASReviewData object from file.

    Parameters
    ----------
    project_id: str, iterable
        The project identifier.
    use_cache: bool
        Use the pickle file if available.
    save_cache: bool
        Save the file to a pickle file if not available.

    Returns
    -------
    ASReviewData:
        The data object for internal use in ASReview.

    """
    project_path = get_project_path(project_id)

    # use cache file
    if use_cache:
        try:
            return _read_data_from_cache(project_id)
        except CacheDataError:
            pass

    # load from file
    fp_data = get_data_file_path(project_path)
    data_obj = ASReviewData.from_file(fp_data)

    # save a pickle version
    if save_cache:
        _write_data_to_cache(project_id, data_obj)

    return data_obj



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sqlite3
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.sparse import load_npz
from scipy.sparse import save_npz
from scipy.sparse.csr import csr_matrix

from asreview._version import get_versions
from asreview.settings import ASReviewSettings
from asreview.state.base import BaseState
from asreview.state.errors import StateError
from asreview.state.errors import StateNotFoundError
from asreview.state.paths import get_feature_matrix_path
from asreview.state.paths import get_project_file_path
from asreview.state.paths import get_settings_metadata_path
from asreview.state.paths import get_sql_path

REQUIRED_TABLES = [
    # the table with the labeling decisions and models trained
    "results",
    # the mapping of record identifiers to row numbers
    "record_table",
    # the latest probabilities.
    "last_probabilities",
    # the latest ranking.
    "last_ranking",
    # the record ids whose labeling decision was changed.
    "decision_changes"
]

RESULTS_TABLE_COLUMNS = [
    "record_id", "label", "classifier", "query_strategy", "balance_strategy",
    "feature_extraction", "training_set", "labeling_time", "notes"
]
SETTINGS_METADATA_KEYS = ["settings", "state_version", "software_version"]


# TODO(State): Update docstring.
class SqlStateV1(BaseState):
    """Class for storing the review state with HDF5 storage.

    Arguments
    ---------
    read_only: bool
        Open state in read only mode. Default False.

    Attributes
    ----------
    settings: asreview.settings.ASReviewSettings
        Return an ASReview settings object with model settings and
        active learning settings.
    n_records_labeled: int
        Get the number of labeled records, where each prior is counted
        individually.
    n_priors: int
        Number of priors. If priors have not been selected returns None.

    """
    def __init__(self, read_only=True):
        super(SqlStateV1, self).__init__(read_only=read_only)

# INTERNAL PATHS AND CONNECTIONS

    def _connect_to_sql(self):
        """Get a connection to the sql database.

        Returns
        -------
        sqlite3.Connection
            Connection to the the sql database.
            The connection is read only if self.read_only is true.
        """
        if self.read_only:
            con = sqlite3.connect(f"file:{str(self._sql_fp)}?mode=ro",
                                  uri=True)
        else:
            con = sqlite3.connect(str(self._sql_fp))
        return con

    # TODO(State): Should this be obtained from webapp/utils/paths, viceversa?
    @property
    def _sql_fp(self):
        """Path to the sql database."""
        return get_sql_path(self.working_dir, self.review_id)

    @property
    def _settings_metadata_fp(self):
        """Path to the settings and metadata json file."""
        return get_settings_metadata_path(self.working_dir, self.review_id)

    @property
    def _feature_matrix_fp(self):
        """Path to the .npz file of the feature matrix"""
        with open(self._settings_metadata_fp, "r") as f:
            feature_extraction = json.load(f)["settings"]["feature_extraction"]

        return get_feature_matrix_path(self.working_dir, feature_extraction)

    def _add_state_file_to_project(self,
                                   review_id,
                                   start_time=None,
                                   review_finished=False):

        if start_time is None:
            start_time = datetime.now()

        # Add the review to the project json.
        with open(get_project_file_path(self.working_dir), "r") as f:
            project_config = json.load(f)

        review_config = {
            "id": review_id,
            "start_time": str(start_time),
            "review_finished": review_finished
        }

        project_config["reviews"].append(review_config)

        with open(get_project_file_path(self.working_dir), "w") as f:
            json.dump(project_config, f)

    def _create_new_state_file(self, working_dir, review_id):
        """Create the files for a new state given an review_id.

        Stages:
        1: create result structure
        2: create model settings
        3: add state to the project file

        Arguments
        ---------
        working_dir: str, pathlib.Path
            Project file location.
        review_id: str
            Identifier (UUID4) of the review.
        """
        if self.read_only:
            raise ValueError("Can't create new state file in read_only mode.")

        self.working_dir = Path(working_dir)
        self.review_id = review_id

        # create folder in the folder `results` with the name of result_id
        self._sql_fp.parent.mkdir(parents=True, exist_ok=True)

        # Create results table.
        con = self._connect_to_sql()
        try:
            cur = con.cursor()

            # Create the results table.
            cur.execute("""CREATE TABLE results
                                (record_id INTEGER,
                                label INTEGER,
                                classifier TEXT,
                                query_strategy TEXT,
                                balance_strategy TEXT,
                                feature_extraction TEXT,
                                training_set INTEGER,
                                labeling_time INTEGER,
                                notes TEXT)""")

            # Create the record table.
            cur.execute("""CREATE TABLE record_table
                                (record_id INT)""")

            # Create the last_probabilities table.
            cur.execute("""CREATE TABLE last_probabilities
                                (proba REAL)""")

            # Create the last_ranking table.
            cur.execute('''CREATE TABLE last_ranking
                                (record_id INTEGER,
                                ranking INT,
                                classifier TEXT,
                                query_strategy TEXT,
                                balance_strategy TEXT,
                                feature_extraction TEXT,
                                training_set INTEGER,
                                time INTEGER)''')

            # Create the table of changed decisions.
            cur.execute('''CREATE TABLE decision_changes
                                (record_id INTEGER,
                                new_label INTEGER,
                                time INTEGER)''')

            con.commit()
            con.close()
        except sqlite3.Error as e:
            con.close()
            raise e

        # Create settings_metadata.json file
        # content of the settings is added later
        self.settings_metadata = {
            "settings": None,
            "state_version": "1",
            "software_version": get_versions()['version']
        }

        with open(self._settings_metadata_fp, "w") as f:
            json.dump(self.settings_metadata, f)

        # after succesfull init, add review_id to the project file
        self._add_state_file_to_project(review_id)

    def _restore(self, working_dir, review_id):
        """
        Initialize a state from files.

        Arguments
        ---------
        working_dir: str, pathlib.Path
            Project file location.
        review_id: str
            Identifier of the review.
        """
        # store filepath
        self.working_dir = Path(working_dir)
        self.review_id = review_id

        # If state already exist
        if not self.working_dir.is_dir():
            raise StateNotFoundError(f"Project {working_dir} doesn't exist.")

        if not self._sql_fp.parent.is_dir():
            raise StateNotFoundError(
                f"Review with id {review_id} doesn't exist.")

        # Cache the settings.
        try:
            with open(self._settings_metadata_fp, "r") as f:
                self.settings_metadata = json.load(f)
        except FileNotFoundError:
            raise AttributeError(
                "'settings_metadata.json' not found in the state file.")

        try:
            if not self._is_valid_version():
                raise ValueError(
                    f"State cannot be read: state version {self.version}, "
                    f"state file version {self.version}.")
        except AttributeError as err:
            raise ValueError(
                f"Unexpected error when opening state file: {err}")

        self._is_valid_state()

    def _is_valid_state(self):
        con = self._connect_to_sql()
        cur = con.cursor()

        # Check if all required tables are present.
        table_names = cur.execute(
            "SELECT name FROM sqlite_master WHERE type='table';").fetchall()
        table_names = [tup[0] for tup in table_names]
        for table in REQUIRED_TABLES:
            if table not in table_names:
                raise StateError(
                    f"The sql file should contain a table named '{table}'.")

        # Check if all required columns are present in results.
        column_names = cur.execute("PRAGMA table_info(results)").fetchall()
        column_names = [tup[1] for tup in column_names]
        for column in RESULTS_TABLE_COLUMNS:
            if column not in column_names:
                raise StateError(
                    f"The results table does not contain the column {column}.")

        # Check settings_metadata contains the required keys.
        settings_metadata_keys = self.settings_metadata.keys()
        for key in SETTINGS_METADATA_KEYS:
            if key not in settings_metadata_keys:
                raise StateError(
                    f"The key {key} was not found in settings_metadata.")

    def close(self):
        pass

# PROPERTIES

    def _is_valid_version(self):
        """Check compatibility of state version."""
        return self.version[0] == "1"

    @property
    def version(self):
        """Version number of the state file.

        Returns
        -------
        str:
            Returns the version of the state file.

        """
        try:
            return self.settings_metadata["state_version"]
        except KeyError:
            raise AttributeError(
                "'settings_metadata.json' does not contain 'state_version'.")

    @property
    def settings(self):
        """Settings of the ASReview pipeline.

        Settings like models.

        Example
        -------

        Example of settings.

            model             : nb
            query_strategy    : max_random
            balance_strategy  : triple
            feature_extraction: tfidf
            n_instances       : 1
            n_queries         : 1
            n_prior_included  : 10
            n_prior_excluded  : 10
            mode              : simulate
            model_param       : {'alpha': 3.822}
            query_param       : {'strategy_1': 'max', 'strategy_2': 'random',
            'mix_ratio': 0.95}
            feature_param     : {}
            balance_param     : {'a': 2.155, 'alpha': 0.94, ... 'gamma': 2.0,
            'shuffle': True}
            abstract_only     : False

        """
        settings = self.settings_metadata['settings']
        if settings is None:
            return None
        return ASReviewSettings(**settings)

    @settings.setter
    def settings(self, settings):
        if isinstance(settings, ASReviewSettings):
            self._add_settings_metadata("settings", settings.to_dict())
        else:
            raise ValueError(
                "'settings' should be an ASReviewSettings object.")

    @property
    def current_queries(self):
        """Get the current queries made by the model.

        This is useful to get back exactly to the state it was in before
        shutting down a review.

        Returns
        -------
        dict:
            The last known queries according to the state file.
        """
        str_queries = self.settings_metadata['current_queries']
        return {int(key): value for key, value in str_queries.items()}

    @current_queries.setter
    def current_queries(self, current_queries):
        str_queries = {
            str(key): value
            for key, value in current_queries.items()
        }
        self._add_settings_metadata("current_queries", str_queries)

    @property
    def n_records_labeled(self):
        _, labeled, _ = self.get_pool_labeled_pending()
        return len(labeled)

    @property
    def n_priors(self):
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute(
            "SELECT COUNT (*) FROM results WHERE query_strategy='prior'")
        n = cur.fetchone()
        con.close()
        n = n[0]

        if n == 0:
            return None
        return n

    @property
    def exist_new_labeled_records(self):
        """Return True if there were records labeled since the last time the
        model training data was saved in the state. Also returns True if no
        model was trained yet, but priors have been added."""
        _, labeled, _ = self.get_pool_labeled_pending()
        last_training_set = self.get_last_ranking()['training_set']
        if last_training_set.empty:
            return len(labeled) > 0
        else:
            return len(labeled) > last_training_set.iloc[0]

    @property
    def model_has_trained(self):
        """Return True if there is data of a trained model in the state."""
        return not self.get_last_ranking().empty

    def _update_project_with_feature_extraction(self, feature_extraction):
        """If the feature extraction method is set, update the project.json."""
        # TODO(State): Should this always be .npz?
        feature_matrix_filename = f'{feature_extraction}_feature_matrix.npz'

        with open(get_project_file_path(self.working_dir), "r") as f:
            project_config = json.load(f)

        # Update the feature matrices section.
        all_matrices = [x['id'] for x in project_config['feature_matrices']]
        if feature_extraction not in all_matrices:
            project_config['feature_matrices'].append({
                'id':
                feature_extraction,
                'filename':
                feature_matrix_filename
            })

        with open(get_project_file_path(self.working_dir), "w") as f:
            json.dump(project_config, f)

    def _add_settings_metadata(self, key, value):
        """Add information to the settings_metadata dictionary."""
        if self.read_only:
            raise ValueError("Can't change settings in read only mode.")
        self.settings_metadata[key] = value

        # If the feature extraction method is being set, update project.json
        if key == 'settings':
            self._update_project_with_feature_extraction(
                value['feature_extraction'])

        with open(self._settings_metadata_fp, "w") as f:
            json.dump(self.settings_metadata, f)

    def add_record_table(self, record_ids):
        # Add the record table to the sql.
        record_sql_input = [(int(record_id), ) for record_id in record_ids]

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("DELETE FROM record_table")
        cur.executemany(
            """INSERT INTO record_table VALUES
                                            (?)""", record_sql_input)
        con.commit()

    def add_last_probabilities(self, probabilities):
        """Save the probabilities of the last model."""
        proba_sql_input = [(proba, ) for proba in probabilities]

        con = self._connect_to_sql()
        cur = con.cursor()

        # Check that the number of rows in the table is 0 (if the table is not
        # yet populated), or that it's equal to len(probabilities).
        cur.execute("SELECT COUNT (*) FROM last_probabilities")
        proba_length = cur.fetchone()[0]
        if not ((proba_length == 0) or (proba_length == len(proba_sql_input))):
            raise ValueError(
                f"There are {proba_length} probabilities in the database, "
                f"but 'probabilities' has length {len(probabilities)}")

        cur.execute("""DELETE FROM last_probabilities""")
        cur.executemany(
            """INSERT INTO last_probabilities VALUES
                                            (?)""", proba_sql_input)
        con.commit()

    def add_last_ranking(self, ranking, classifier, query_strategy,
                         balance_strategy, feature_extraction, training_set):
        """Save the ranking of the last iteration of the model."""
        record_ids = self.get_record_table()

        if len(record_ids) != len(ranking):
            raise ValueError("The ranking should have the same length as the "
                             "record table.")

        classifiers = [classifier for _ in record_ids]
        query_strategies = [query_strategy for _ in record_ids]
        balance_strategies = [balance_strategy for _ in record_ids]
        feature_extractions = [feature_extraction for _ in record_ids]
        training_sets = [int(training_set) for _ in record_ids]
        ranking_times = [datetime.now()] * len(record_ids)

        # Create the database rows.
        db_rows = [(int(record_ids[i]), int(ranking[i]), classifiers[i],
                    query_strategies[i], balance_strategies[i],
                    feature_extractions[i], training_sets[i], ranking_times[i])
                   for i in range(len(record_ids))]

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("DELETE FROM last_ranking")
        cur.executemany(
            """INSERT INTO last_ranking VALUES
                                    (?, ?, ?, ?, ?, ?, ?, ?)""", db_rows)
        con.commit()
        con.close()

    def add_feature_matrix(self, feature_matrix):
        """Add feature matrix to project file.

        Feature matrices are stored in the project file. See
        asreview.state.SqlStateV1._feature_matrix_fp for the file
        location.

        Arguments
        ---------
        feature_matrix: numpy.ndarray, scipy.sparse.csr.csr_matrix
            The feature matrix to add to the project file.
        """
        # Make sure the feature matrix is in csr format.
        if isinstance(feature_matrix, np.ndarray):
            feature_matrix = csr_matrix(feature_matrix)
        if not isinstance(feature_matrix, csr_matrix):
            raise ValueError(
                "The feature matrix should be convertible to type "
                "scipy.sparse.csr.csr_matrix.")

        save_npz(self._feature_matrix_fp, feature_matrix)

    def get_feature_matrix(self):
        """Get the feature matrix from the project file.

        Returns
        -------
        numpy.ndarray:
            Returns the feature matrix.
        """
        return load_npz(self._feature_matrix_fp)

    def add_note(self, note, record_id):
        """Add a text note to save with a labeled record."""
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("UPDATE results SET notes = ? WHERE record_id = ?",
                    (note, record_id))
        con.commit()
        con.close()

    def add_labeling_data(self, record_ids, labels, notes=None, prior=False):
        """Add all the data of one labeling action."""

        # TODO (State): Add custom datasets.
        # TODO (State): Add models being trained.

        # Check if the state is still valid.
        self._is_valid_state()

        labeling_times = [datetime.now()] * len(record_ids)

        if notes is None:
            notes = [None for _ in record_ids]

        lengths = [len(record_ids), len(labels), len(notes)]
        # Check that all input data has the same length.
        if len(set(lengths)) != 1:
            raise ValueError("Input data should be of the same length.")
        n_records_labeled = len(record_ids)

        pool, _, pending = self.get_pool_labeled_pending()

        if prior:
            # Check that the record_ids are in the pool.
            if not all(record_id in pool.values for record_id in record_ids):
                raise ValueError("Labeling priors, but not all "
                                 "record_ids were found in the pool.")

            query_strategies = ['prior' for _ in record_ids]
            training_sets = [-1 for _ in record_ids]
            data = [(int(record_ids[i]), int(labels[i]), query_strategies[i],
                     training_sets[i], labeling_times[i], notes[i])
                    for i in range(n_records_labeled)]

            # If prior, we need to insert new records into the database.
            query = ("INSERT INTO results (record_id, label, query_strategy, "
                     "training_set, labeling_time, notes) "
                     "VALUES (?, ?, ?, ?, ?, ?)")

        else:
            # Check that the record_ids are pending.
            if not all(record_id in pending.values
                       for record_id in record_ids):
                raise ValueError("Labeling records, but not all "
                                 "record_ids were pending.")

            data = [(int(labels[i]), labeling_times[i], notes[i],
                     int(record_ids[i])) for i in range(n_records_labeled)]

            # If not prior, we need to update records.
            query = ("UPDATE results SET label=?, labeling_time=?, "
                     "notes=? WHERE record_id=?")

        # Add the rows to the database.
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.executemany(query, data)
        con.commit()
        con.close()

    def change_decision(self, record_id):
        """Change the label of a record from 0 to 1 or vice versa."""
        current_label = self.get_data_by_record_id(record_id)['label'][0]

        # New label should be of type int, not np.int, to insert into sql.
        new_label = int(1 - current_label)
        current_time = datetime.now()

        con = self._connect_to_sql()
        cur = con.cursor()

        # Change the label.
        cur.execute("UPDATE results SET label = ? WHERE record_id = ?",
                    (new_label, record_id))

        # Add the change to the decision changes table.
        cur.execute("INSERT INTO decision_changes VALUES (?,?, ?)",
                    (record_id, new_label, current_time))

        con.commit()
        con.close()

    def delete_record_labeling_data(self, record_id):
        """Delete the labeling data for the given record id."""
        current_time = datetime.now()

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute('DELETE FROM results WHERE record_id=?', (record_id, ))

        # Add the change to the decision changes table.
        cur.execute("INSERT INTO decision_changes VALUES (?,?, ?)",
                    (record_id, None, current_time))
        con.commit()
        con.close()

    def get_decision_changes(self):
        """Get the record ids of the records whose labels have been changed
        after the original labeling action."""
        con = self._connect_to_sql()
        change_table = pd.read_sql_query('SELECT * FROM decision_changes', con)
        con.close()
        return change_table

    def get_record_table(self):
        """Get the record table of the state file.

        Returns
        -------
        pd.Series:
            Series with name 'record_id' containing the record ids.
        """
        con = self._connect_to_sql()
        record_table = pd.read_sql_query('SELECT * FROM record_table', con)
        record_table = record_table['record_id']
        con.close()
        return record_table

    def get_pool_labeled_pending(self):
        """Return the labeled and unlabeled records and the records pending a
        labeling decision.

        Returns
        -------
        tuple (pd.Series, pd.DataFrame, pd.Series):
            Returns a tuple (pool, labeled, pending). Pool is a series
            containing the unlabeled, not pending record_ids, ordered by the
            last predicted ranking of the model. Labeled is a dataframe
            containing the record_ids and labels of the labeled records, in the
            order that they were labeled. Pending is a series containing the
            record_ids of the records whose label is pending.
        """
        con = self._connect_to_sql()

        query = """SELECT record_table.record_id, results.label,
                results.rowid AS label_order, results.classifier,
                last_ranking.ranking
                FROM record_table
                LEFT JOIN results
                ON results.record_id=record_table.record_id
                LEFT JOIN last_ranking
                ON record_table.rowid=last_ranking.rowid
                """

        df = pd.read_sql_query(query, con)
        con.close()
        labeled = df.loc[~df['label'].isna()] \
            .sort_values('label_order') \
            .loc[:, ['record_id', 'label']] \
            .astype(int)
        pool = df.loc[df['label_order'].isna()] \
            .sort_values('ranking') \
            .loc[:, 'record_id'] \
            .astype(int)
        pending = df.loc[df['label'].isna() & ~df['classifier'].isna()] \
            .sort_values('label_order') \
            .loc[:, 'record_id'] \
            .astype(int)

        return pool, labeled, pending

    def get_last_probabilities(self):
        """Get the probabilities produced by the last classifier.

        Returns
        -------
        pd.DataFrame:
            Dataframe with column 'proba' containing the probabilities.
        """
        con = self._connect_to_sql()
        last_probabilities = pd.read_sql_query(
            'SELECT * FROM last_probabilities', con)
        con.close()
        return last_probabilities

    def get_last_ranking(self):
        """Get the ranking from the state."""
        con = self._connect_to_sql()
        last_ranking = pd.read_sql_query('SELECT * FROM last_ranking', con)
        con.close()
        return last_ranking

    def _move_ranking_data_to_results(self, record_ids):
        """Move the data with the given record_ids from the last_ranking table
        to the results table.

        Arguments
        ---------
        record_ids: list
            List of record ids in last ranking whose model data should be added
            to the results table.
        """
        if self.model_has_trained:
            record_list = [(record_id, ) for record_id in record_ids]
            con = self._connect_to_sql()
            cur = con.cursor()
            cur.executemany(
                """INSERT INTO results (record_id, classifier, query_strategy,
                balance_strategy, feature_extraction, training_set)
                SELECT record_id, classifier, query_strategy,
                balance_strategy, feature_extraction, training_set
                FROM last_ranking
                WHERE record_id=?""", record_list)
            con.commit()
            con.close()
        else:
            raise StateError("Save trained model data "
                             "before using this function.")

    def query_top_ranked(self, n):
        """Get the top n instances from the pool according to the last ranking.
        Add the model data to the results table.

        Arguments
        ---------
        n: int
            Number of instances.

        Returns
        -------
        list
            List of record_ids of the top n ranked records.
        """
        if self.model_has_trained:
            pool, _, _ = self.get_pool_labeled_pending()
            top_n_records = pool[:n].to_list()
            self._move_ranking_data_to_results(top_n_records)
        else:
            raise StateError("Save trained model data "
                             "before using this function.")

        return top_n_records

    def get_data_by_query_number(self, query, columns=None):
        """Get the data of a specific query from the results table.

        Arguments
        ---------
        query: int
            Number of the query of which you want the data. query=0 corresponds
            to all the prior records.
        columns: list
            List of columns names of the results table.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data from the results table with the given
            query number and columns.
        """
        if columns is not None:
            if not type(columns) == list:
                raise ValueError("The columns argument should be a list.")
        col_query_string = '*' if columns is None else ','.join(columns)

        if query == 0:
            sql_query = f"SELECT {col_query_string} FROM results WHERE " \
                        f"query_strategy='prior'"
        else:
            rowid = query + self.n_priors
            sql_query = f"SELECT {col_query_string} FROM results WHERE " \
                        f"rowid={rowid}"

        con = self._connect_to_sql()
        data = pd.read_sql_query(sql_query, con)
        con.close()
        return data

    def get_data_by_record_id(self, record_id, columns=None):
        """Get the data of a specific query from the results table.

        Arguments
        ---------
        record_id: int
            Record id of which you want the data.
        columns: list
            List of columns names of the results table.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data from the results table with the given
            record_id and columns.
        """
        query_string = '*' if columns is None else ','.join(columns)

        con = self._connect_to_sql()
        data = pd.read_sql_query(
            f'SELECT {query_string} FROM results WHERE record_id={record_id}',
            con)
        con.close()
        return data

    def get_dataset(self, columns=None):
        """Get a column from the results table.

        Arguments
        ---------
        columns: list, str
            List of columns names of the results table, or a string containing
            one column name.

        Returns
        -------
        pd.DataFrame:
            Dataframe containing the data of the specified columns of the
            results table.
        """
        if type(columns) == str:
            columns = [columns]
        query_string = '*' if columns is None else ','.join(columns)
        con = self._connect_to_sql()
        data = pd.read_sql_query(f'SELECT {query_string} FROM results', con)
        con.close()
        return data

    def get_order_of_labeling(self):
        """Get full array of record id's in order that they were labeled.

        Returns
        -------
        pd.Series:
            The record_id's in the order that they were labeled.
        """
        return self.get_dataset('record_id')['record_id']

    def get_priors(self):
        """Get the record ids of the priors.

        Returns
        -------
        pd.Series:
            The record_id's of the priors in the order they were added.
        """
        return self.get_order_of_labeling()[:self.n_priors]

    def get_labels(self):
        """Get the labels from the state file.

        Returns
        -------
        pd.Series:
            Series containing the labels at each labelling moment.
        """
        return self.get_dataset('label')['label'].dropna()

    def get_classifiers(self):
        """Get the classifiers from the state file.

        Returns
        -------
        pd.Series:
            Series containing the classifier used at each labeling moment.
        """
        return self.get_dataset('classifier')['classifier']

    def get_query_strategies(self):
        """Get the query strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the query strategy used to get the record to
            query at each labeling moment.
        """
        return self.get_dataset('query_strategy')['query_strategy']

    def get_balance_strategies(self):
        """Get the balance strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the balance strategy used to get the training
            data at each labeling moment.
        """
        return self.get_dataset('balance_strategy')['balance_strategy']

    def get_feature_extraction(self):
        """Get the query strategies from the state file.

        Returns
        -------
        pd.Series:
            Series containing the feature extraction method used for the
            classifier input at each labeling moment.
        """
        return self.get_dataset('feature_extraction')['feature_extraction']

    def get_training_sets(self):
        """Get the training_sets from the state file.

        Returns
        -------
        pd.Series:
            Series containing the training set on which the classifier was fit
            at each labeling moment.
        """
        return self.get_dataset('training_set')['training_set']

    def get_labeling_times(self, time_format='int'):
        """Get the time of labeling the state file.

        Arguments
        ---------
        time_format: 'int' or 'datetime'
            Format of the return value. If it is 'int' you get a UTC timestamp,
            if it is 'datetime' you get datetime instead of an integer.

        Returns
        -------
        pd.Series:
            If format='int' you get a UTC timestamp (integer number of
            microseconds), if it is 'datetime' you get datetime format.
        """
        times = self.get_dataset('labeling_time')['labeling_time'].dropna()

        # Convert time to datetime format.
        if time_format == 'datetime':
            times = times.applymap(
                lambda x: datetime.utcfromtimestamp(x / 10**6))

        return times



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import logging
from datetime import datetime
from pathlib import Path
from pathlib import PurePath

import numpy as np

from asreview.compat import convert_id_to_idx
from asreview.config import ASCII_LOGO
from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_FEATURE_EXTRACTION
from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_N_INSTANCES
from asreview.config import DEFAULT_N_PRIOR_EXCLUDED
from asreview.config import DEFAULT_N_PRIOR_INCLUDED
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import EMAIL_ADDRESS
from asreview.config import GITHUB_PAGE
from asreview.config import KERAS_MODELS
from asreview.config import LABEL_NA
from asreview.data import ASReviewData
from asreview.data import load_data
from asreview.entry_points.base import BaseEntryPoint
from asreview.entry_points.base import _base_parser
from asreview.io.paper_record import preview_record
from asreview.models.balance.utils import get_balance_model
from asreview.models.classifiers import get_classifier
from asreview.models.feature_extraction import get_feature_model
from asreview.models.query import get_query_model
from asreview.review.simulate import ReviewSimulate
from asreview.settings import ASReviewSettings
from asreview.state.paths import get_data_path
from asreview.state.paths import get_project_file_path
from asreview.state.utils import init_project_folder_structure
from asreview.types import type_n_queries
from asreview.utils import get_random_state

ASCII_MSG_SIMULATE = """
---------------------------------------------------------------------------------
|                                                                                |
|  Welcome to the ASReview Automated Systematic Review software.                 |
|  In this mode the computer will simulate how well the ASReview software        |
|  could have accelerate the systematic review of your dataset.                  |
|  You can sit back and relax while the computer runs this simulation.           |
|                                                                                |
|  GitHub page:        {0: <58}|
|  Questions/remarks:  {1: <58}|
|                                                                                |
---------------------------------------------------------------------------------
""".format(GITHUB_PAGE, EMAIL_ADDRESS)  # noqa


def review_finished(project_path, review_id=None):
    """Mark a review in the project as finished. If no review_id is given,
    mark the first review as finished.

    Arguments
    ---------
    project_path: pathlike
        Path to the project folder.
    review_id: str
        Identifier of the review to mark as finished.
    """
    project_path = Path(project_path)
    with open(get_project_file_path(project_path), 'r') as f:
        project_config = json.load(f)

    if review_id is None:
        review_index = 0
    else:
        review_index = [x['id']
                        for x in project_config['reviews']].index(review_id)

    project_config['reviews'][review_index]['review_finished'] = True
    project_config['reviews'][review_index]['end_time'] = str(datetime.now())

    with open(get_project_file_path(project_path), 'w') as f:
        json.dump(project_config, f)


def _set_log_verbosity(verbose):
    if verbose == 0:
        logging.getLogger().setLevel(logging.WARNING)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose >= 2:
        logging.getLogger().setLevel(logging.DEBUG)


class SimulateEntryPoint(BaseEntryPoint):
    description = "Simulate the performance of ASReview."

    def execute(self, argv):

        # parse arguments
        parser = _simulate_parser()
        args = parser.parse_args(argv)

        # change the verbosity
        _set_log_verbosity(args.verbose)

        # check for state file extension
        if args.state_file is None:
            raise ValueError(
                "Specify project file name (with .asreview extension).")

        # print intro message
        print(ASCII_LOGO + ASCII_MSG_SIMULATE)

        # read dataset into memory
        as_data = load_data(args.dataset)

        if len(as_data) == 0:
            raise ValueError("Supply at least one dataset"
                             " with at least one record.")

        # rewrite to ProjectAPI
        init_project_folder_structure(args.state_file, project_mode='simulate')

        # Add the dataset to the project file.
        as_data.to_csv(Path(get_data_path(args.state_file), f'{as_data.data_name}.csv'))

        # create a new settings object from arguments
        settings = ASReviewSettings(model=args.model,
                                    n_instances=args.n_instances,
                                    n_queries=args.n_queries,
                                    n_papers=args.n_papers,
                                    n_prior_included=args.n_prior_included,
                                    n_prior_excluded=args.n_prior_excluded,
                                    query_strategy=args.query_strategy,
                                    balance_strategy=args.balance_strategy,
                                    feature_extraction=args.feature_extraction,
                                    mode="simulate",
                                    data_fp=None)
        settings.from_file(args.config_file)

        # Initialize models.
        random_state = get_random_state(args.seed)
        train_model = get_classifier(settings.model,
                                     **settings.model_param,
                                     random_state=random_state)
        query_model = get_query_model(settings.query_strategy,
                                      **settings.query_param,
                                      random_state=random_state)
        balance_model = get_balance_model(settings.balance_strategy,
                                          **settings.balance_param,
                                          random_state=random_state)
        feature_model = get_feature_model(settings.feature_extraction,
                                          **settings.feature_param,
                                          random_state=random_state)

        # TODO{Deprecate and integrate with the model}
        # LSTM models need embedding matrices.
        if train_model.name.startswith("lstm-"):
            texts = as_data.texts
            train_model.embedding_matrix = feature_model.get_embedding_matrix(
                texts, args.embedding_fp)

        # prior knowledge
        if args.prior_idx is not None and args.prior_record_id is not None and \
                len(args.prior_idx) > 0 and len(args.prior_record_id) > 0:
            raise ValueError(
                "Not possible to provide both prior_idx and prior_record_id")

        prior_idx = args.prior_idx
        if args.prior_record_id is not None and len(args.prior_record_id) > 0:
            prior_idx = convert_id_to_idx(as_data, args.prior_record_id)

        print("The following records are prior knowledge:\n")
        for prior_record_id in args.prior_record_id:
            preview = preview_record(as_data.record(prior_record_id))
            print(f"{prior_record_id} - {preview}")

        # Initialize the review class.
        reviewer = ReviewSimulate(as_data,
                                  state_file=args.state_file,
                                  model=train_model,
                                  query_model=query_model,
                                  balance_model=balance_model,
                                  feature_model=feature_model,
                                  n_papers=args.n_papers,
                                  n_instances=args.n_instances,
                                  n_queries=args.n_queries,
                                  prior_indices=prior_idx,
                                  n_prior_included=args.n_prior_included,
                                  n_prior_excluded=args.n_prior_excluded,
                                  init_seed=args.init_seed)

        # Start the review process.
        reviewer.review()

        # Mark review as finished.
        review_finished(args.state_file)


DESCRIPTION_SIMULATE = """
ASReview for simulation.

The simulation modus is used to measure the performance of our
software on existing systematic reviews. The software shows how many
papers you could have potentially skipped during the systematic
review."""


def _simulate_parser(prog="simulate", description=DESCRIPTION_SIMULATE):
    parser = _base_parser(prog=prog, description=description)
    # Active learning parameters
    # File path to the data.
    parser.add_argument(
        "dataset",
        type=str,
        help="File path to the dataset or one of the benchmark datasets."
    )
    # Initial data (prior knowledge)
    parser.add_argument(
        "--n_prior_included",
        default=DEFAULT_N_PRIOR_INCLUDED,
        type=int,
        help="Sample n prior included papers. "
             "Only used when --prior_idx is not given. "
             f"Default {DEFAULT_N_PRIOR_INCLUDED}")

    parser.add_argument(
        "--n_prior_excluded",
        default=DEFAULT_N_PRIOR_EXCLUDED,
        type=int,
        help="Sample n prior excluded papers. "
             "Only used when --prior_idx is not given. "
             f"Default {DEFAULT_N_PRIOR_EXCLUDED}")

    parser.add_argument(
        "--prior_idx",
        default=[],
        nargs="*",
        type=int,
        help="Prior indices by rownumber (0 is first rownumber)."
    )
    parser.add_argument(
        "--prior_record_id",
        default=[],
        nargs="*",
        type=int,
        help="Prior indices by record_id."
    )
    # logging and verbosity
    parser.add_argument(
        "--state_file", "-s",
        default=None,
        type=str,
        help="Location to store the state of the simulation."
    )
    parser.add_argument(
        "-m", "--model",
        type=str,
        default=DEFAULT_MODEL,
        help=f"The prediction model for Active Learning. "
             f"Default: '{DEFAULT_MODEL}'.")
    parser.add_argument(
        "-q", "--query_strategy",
        type=str,
        default=DEFAULT_QUERY_STRATEGY,
        help=f"The query strategy for Active Learning. "
             f"Default: '{DEFAULT_QUERY_STRATEGY}'.")
    parser.add_argument(
        "-b", "--balance_strategy",
        type=str,
        default=DEFAULT_BALANCE_STRATEGY,
        help="Data rebalancing strategy mainly for RNN methods. Helps against"
             " imbalanced dataset with few inclusions and many exclusions. "
             f"Default: '{DEFAULT_BALANCE_STRATEGY}'")
    parser.add_argument(
        "-e", "--feature_extraction",
        type=str,
        default=DEFAULT_FEATURE_EXTRACTION,
        help="Feature extraction method. Some combinations of feature"
             " extraction method and prediction model are impossible/ill"
             " advised."
             f"Default: '{DEFAULT_FEATURE_EXTRACTION}'"
    )
    parser.add_argument(
        '--init_seed',
        default=None,
        type=int,
        help="Seed for setting the prior indices if the --prior_idx option is "
             "not used. If the option --prior_idx is used with one or more "
             "index, this option is ignored."
    )
    parser.add_argument(
        "--n_instances",
        default=DEFAULT_N_INSTANCES,
        type=int,
        help="Number of papers queried each query."
             f"Default {DEFAULT_N_INSTANCES}.")
    parser.add_argument(
        "--n_queries",
        type=type_n_queries,
        default=None,
        help="The number of queries. Alternatively, entering 'min' will stop the "
             "simulation when all relevant records have been found. By default, "
             "the program stops after all records are reviewed or is interrupted "
             "by the user."
    )
    parser.add_argument(
        "-n", "--n_papers",
        type=int,
        default=None,
        help="The number of papers to be reviewed. By default, "
             "the program stops after all documents are reviewed or is "
             "interrupted by the user."
    )
    parser.add_argument(
        "--verbose", "-v",
        default=0,
        type=int,
        help="Verbosity"
    )

    return parser



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import zipfile
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

from asreview import __version__ as asreview_version
from asreview.config import LABEL_NA
from asreview.config import PROJECT_MODES
from asreview.state.errors import StateError
from asreview.state.errors import StateNotFoundError
from asreview.state.paths import get_data_file_path
from asreview.state.paths import get_data_path
from asreview.state.paths import get_labeled_path
from asreview.state.paths import get_lock_path
from asreview.state.paths import get_pool_path
from asreview.state.paths import get_project_file_path
from asreview.state.paths import get_reviews_path
from asreview.state.paths import get_state_path
from asreview.state.paths import get_tmp_path
from asreview.state.utils import delete_state_from_project
from asreview.state.utils import init_project_folder_structure
from asreview.state.utils import open_state
from asreview.webapp.sqlock import SQLiteLock
from asreview.webapp.utils.io import read_data
from asreview.webapp.utils.project_path import asreview_path
from asreview.webapp.utils.project_path import get_project_path
from asreview.webapp.utils.project_path import list_asreview_project_paths
from asreview.webapp.utils.validation import is_project
from asreview.webapp.utils.validation import is_v0_project


class ProjectNotFoundError(Exception):
    pass


def _get_executable():
    """Get the Python executable"""

    py_exe = sys.executable

    # if the exe is an empty string or None, then fall badck on 'python'
    if not py_exe:
        py_exe = 'python'

    return py_exe


def _create_project_id(name):
    """Create project id from input name."""

    if isinstance(name, str) \
            and len(name) > 0 \
            and not name[0].isalnum():
        raise ValueError(
            "First character should be alphabet"
            " letter (a-z) or number (0-9).")

    if not name \
            and not isinstance(name, str) \
            and len(name) >= 3:
        raise ValueError(
            "Project name should be at least 3 characters.")

    project_id = ""
    for c in name.lower():
        if c.isalnum():
            project_id += c
        elif len(project_id) > 0 and project_id[-1] != "-":
            project_id += "-"

    return project_id


def init_project(project_id,
                 project_mode="oracle",
                 project_name=None,
                 project_description=None,
                 project_authors=None):
    """Initialize the necessary files specific to the web app."""
    project_path = get_project_path(project_id)

    if is_project(project_id):
        raise ValueError("Project already exists.")

    if project_mode not in PROJECT_MODES:
        raise ValueError(f"Project mode '{project_mode}' is not in "
                         f"{PROJECT_MODES}.")

    project_config = \
        init_project_folder_structure(project_path,
                                      project_mode=project_mode,
                                      project_name=project_name,
                                      project_description=project_description,
                                      project_authors=project_authors)

    return project_config


def rename_project(project_id, project_name_new):
    """Rename a project id.

    This function only works for projects in ASReview LAB  web interface.
    This is the result of the file storage in
    asreview.webapp.utils.project_path.asreview_path.

    Arguments
    ---------
    project_id: str
        The current project_id.
    project_name_new: str
        The new project name to be converted into a new
        project_id.

    Returns
    -------
    str:
        The new project_id.
    """

    # create a new project_id from project name
    project_id_new = _create_project_id(project_name_new)

    if (project_id == project_id_new):
        # nothing to do
        return project_id

    if (project_id != project_id_new) & is_project(project_id_new):
        raise ValueError(f"Project '{project_id_new}' already exists.")

    project_path = get_project_path(project_id)
    project_path_new = Path(asreview_path(), project_id_new)
    project_file_path_new = get_project_file_path(project_path_new)

    try:
        project_path.rename(project_path_new)

        with open(project_file_path_new, "r") as fp:
            project_info = json.load(fp)

        project_info["id"] = project_id_new
        project_info["name"] = project_name_new

        with open(project_file_path_new, "w") as fp:
            json.dump(project_info, fp)

    except Exception as err:
        raise err

    return project_id_new


def update_project_info(project_id, **kwargs):
    '''Update project info'''

    kwargs_copy = kwargs.copy()

    if "name" in kwargs_copy:
        del kwargs_copy["name"]
        logging.info(
            "Update project name is ignored, use 'rename_project' function.")

    # validate schema
    if "mode" in kwargs_copy and kwargs_copy["mode"] not in PROJECT_MODES:
        raise ValueError(
            "Project mode '{}' not found.".format(kwargs_copy["mode"]))

    # update project file
    project_path = get_project_path(project_id)
    project_file_path = get_project_file_path(project_path)

    with open(project_file_path, "r") as fp:
        project_info = json.load(fp)

    project_info.update(kwargs_copy)

    with open(project_file_path, "w") as fp:
        json.dump(project_info, fp)

    return project_id


def import_project_file(file_name):
    """Import .asreview project file"""

    try:
        # Unzip the project file
        with zipfile.ZipFile(file_name, "r") as zip_obj:
            zip_filenames = zip_obj.namelist()

            # raise error if no ASReview project file
            if "project.json" not in zip_filenames:
                raise ValueError("File doesn't contain valid project format.")

            # extract all files to a temporary folder
            tmpdir = tempfile.mkdtemp()
            zip_obj.extractall(path=tmpdir)

    except zipfile.BadZipFile:
        raise ValueError("File is not an ASReview file.")

    try:
        # Open the project file and check the id. The id needs to be
        # unique, otherwise it is exended with -copy.
        import_project = None
        fp = Path(tmpdir, "project.json")
        with open(fp, "r+") as f:

            # load the project info in scope of function
            import_project = json.load(f)

            # If the uploaded project already exists,
            # then overwrite project.json with a copy suffix.
            while is_project(import_project["id"]):
                # project update
                import_project["id"] = f"{import_project['id']}-copy"
                import_project["name"] = f"{import_project['name']} copy"
            else:
                # write to file
                f.seek(0)
                json.dump(import_project, f)
                f.truncate()

        # location to copy file to
        fp_copy = get_project_path(import_project["id"])
        # Move the project from the temp folder to the projects folder.
        os.replace(tmpdir, fp_copy)

    except Exception:
        # Unknown error.
        raise ValueError("Failed to import project "
                         f"'{file_name.filename}'.")

    return import_project["id"]


def add_dataset_to_project(project_id, file_name):
    """Add file path to the project file.

    Add file to data subfolder and fill the pool of iteration 0.
    """
    project_path = get_project_path(project_id)

    update_project_info(project_id, dataset_path=file_name)

    # fill the pool of the first iteration
    as_data = read_data(project_id)

    state_file = get_state_path(project_path)

    with open_state(state_file, read_only=False) as state:

        # save the record ids in the state file
        state.add_record_table(as_data.record_ids)

        # if the data contains labels, add them to the state file
        if as_data.labels is not None:

            labeled_indices = np.where(as_data.labels != LABEL_NA)[0]
            labels = as_data.labels[labeled_indices].tolist()
            labeled_record_ids = as_data.record_ids[labeled_indices].tolist()

            # add the labels as prior data
            state.add_labeling_data(
                record_ids=labeled_record_ids,
                labels=labels,
                notes=[None for _ in labeled_record_ids],
                prior=True
            )


def remove_dataset_from_project(project_id):
    """Remove dataset from project.

    """
    project_path = get_project_path(project_id)

    # reset dataset_path
    update_project_info(project_id, dataset_path=None)

    # remove datasets from project
    shutil.rmtree(get_data_path(project_path))

    # remove state file if present
    if get_reviews_path(project_path).is_dir() and \
            any(get_reviews_path(project_path).iterdir()):
        delete_state_from_project(project_path)


def add_review_to_project(project_id, simulation_id):
    update_review_in_project(project_id, simulation_id, True)


def update_review_in_project(project_id, review_id, review_finished):
    project_path = get_project_path(project_id)

    # read the file with project info
    with open(get_project_file_path(project_path), "r") as fp:
        project_info = json.load(fp)

    if "reviews" not in project_info:
        project_info["reviews"] = []

    review = {"id": review_id, "state": review_finished}

    project_info["reviews"].append(review)

    # update the file with project info
    with open(get_project_file_path(project_path), "w") as fp:
        json.dump(project_info, fp)


def get_project_config(project_id):
    project_path = get_project_path(project_id)

    try:

        # read the file with project info
        with open(get_project_file_path(project_path), "r") as fp:

            project_info = json.load(fp)

    except FileNotFoundError:
        raise ProjectNotFoundError(f"Project '{project_id}' not found")

    return project_info


def clean_project_tmp_files(project_id):
    """Clean temporary files in a project.

    Arguments
    ---------
    project_id: str
        The id of the current project.
    """
    project_path = get_project_path(project_id)

    # clean pickle files
    for f_pickle in project_path.rglob("*.pickle"):
        try:
            os.remove(f_pickle)
        except OSError as e:
            print(f"Error: {f_pickle} : {e.strerror}")

    # clean tmp export files
    # TODO


def clean_all_project_tmp_files():
    """Clean temporary files in all projects.
    """
    for project_path in list_asreview_project_paths():
        clean_project_tmp_files(project_path)


def get_paper_data(project_id, paper_id, return_debug_label=False):
    """Get the title/authors/abstract for a paper."""
    as_data = read_data(project_id)
    record = as_data.record(int(paper_id), by_index=False)

    paper_data = {}
    paper_data['title'] = record.title
    paper_data['authors'] = record.authors
    paper_data['abstract'] = record.abstract
    paper_data['doi'] = record.doi

    # return the debug label
    debug_label = record.extra_fields.get("debug_label", None)
    paper_data['_debug_label'] = \
        int(debug_label) if pd.notnull(debug_label) else None

    return paper_data


def get_instance(project_id):
    """Get a new instance to review.

    Arguments
    ---------
    project_id: str
        The id of the current project.
    """
    project_path = get_project_path(project_id)
    with open_state(project_path, read_only=False) as state:
        # First check if there is a pending record.
        _, _, pending = state.get_pool_labeled_pending()
        if not pending.empty:
            record_ids = pending.to_list()
        # Else query for a new record.
        else:
            record_ids = state.query_top_ranked(1)

    if len(record_ids) > 0:
        return record_ids[0]
    else:
        # end of pool
        return None


def get_legacy_statistics(json_fp):
    """Return the data necessary for get_statistics from and old json state
    file.

    Arguments
    ---------
    json_fp: pathlike
        File path of the json state file.
    """
    with open(json_fp, 'r') as f:
        s = json.load(f)

    # Get the labels.
    labeled = np.array([
        int(sample_data[1])
        for query in range(len(s['results']))
        for sample_data in s['results'][query]['labelled']
    ])

    # Get the record table.
    data_hash = list(s['data_properties'].keys())[0]
    record_table = s['data_properties'][data_hash][
        'record_table']

    n_records = len(record_table)
    return labeled, n_records


def get_statistics(project_id):
    """Get statistics from project files.

    Arguments
    ---------
    project_id: str
        The id of the current project.

    Returns
    -------
    dict:
        Dictionary with statistics.
    """
    project_path = get_project_path(project_id)

    if is_v0_project(project_id):
        json_fp = Path(project_path, 'result.json')
        # Check if the v0 project is in review.
        if json_fp.exists():
            labels, n_records = get_legacy_statistics(json_fp)
        # No result found.
        else:
            labels = np.array([])
            n_records = 0
    else:
        # Check if there is a review started in the project.
        try:
            with open_state(project_path) as s:
                labels = s.get_labels()
                n_records = len(s.get_record_table())
        # No state file found or not init.
        except (StateNotFoundError, StateError):
            labels = np.array([])
            n_records = 0

    n_included = sum(labels == 1)
    n_excluded = sum(labels == 0)

    if len(labels) > 0:
        n_since_last_relevant = labels.tolist()[::-1].index(1)
    else:
        n_since_last_relevant = 0

    return {
        "n_included": n_included,
        "n_excluded": n_excluded,
        "n_since_last_inclusion": n_since_last_relevant,
        "n_papers": n_records,
        "n_pool": n_records - n_excluded - n_included
    }


def export_to_string(project_id, export_type="csv"):
    project_path = get_project_path(project_id)

    # read the dataset into a ASReview data object
    as_data = read_data(project_id)

    with open_state(project_path) as s:
        proba = s.get_last_probabilities()
        labeled_data = s.get_dataset(['record_id', 'label'])
        record_table = s.get_record_table()

    prob_df = pd.concat([record_table, proba], axis=1)

    ranking = pd. \
        merge(prob_df, labeled_data, on='record_id', how='left'). \
        fillna(0.5). \
        sort_values(['label', 'proba'], ascending=False)['record_id']

    labeled = labeled_data.values.tolist()

    # export the data to file
    if export_type == "csv":
        return as_data.to_csv(fp=None, labels=labeled, ranking=ranking)

    elif export_type == "tsv":
        return as_data.to_csv(
            fp=None, sep="\t", labels=labeled, ranking=ranking)

    elif export_type == "excel":
        get_tmp_path(project_path).mkdir(exist_ok=True)
        fp_tmp_export = Path(get_tmp_path(project_path), "export_result.xlsx")
        return as_data.to_excel(
            fp=fp_tmp_export, labels=labeled, ranking=ranking)

    elif export_type == "ris":
        return as_data.to_ris(fp=None, labels=labeled, ranking=ranking)

    else:
        raise ValueError("This export type isn't implemented.")


def train_model(project_id):
    py_exe = _get_executable()
    run_command = [py_exe, "-m", "asreview", "web_run_model", project_id]
    subprocess.Popen(run_command)


def update_instance(project_id, paper_i, label, retrain_model=True):
    """Update a labeling decision."""
    project_path = get_project_path(project_id)
    state_path = get_state_path(project_path)

    record_id = int(paper_i)
    label = int(label)

    with open_state(state_path, read_only=False) as state:
        record_info = state.get_data_by_record_id(record_id)

        # Check if the record is actually labeled.
        if record_info.empty:
            raise ValueError(f"Tried to update record_id {record_id}, "
                             f"but it has not been labeled yet.")
        else:
            # If the current label is the same as the updated label, do nothing.
            current_label = record_info['label'][0]
            if current_label == label:
                pass
            # Else change the labeling decision.
            else:
                state.change_decision(record_id)

    if retrain_model:
        train_model(project_id)


def label_instance(project_id, paper_i, label, prior=False, retrain_model=True):
    """Label a paper after reviewing the abstract.

    """
    project_path = get_project_path(project_id)
    state_path = get_state_path(project_path)

    paper_i = int(paper_i)
    label = int(label)

    with open_state(state_path, read_only=False) as state:

        # get the index of the active iteration
        if label in [0, 1]:

            # add the labels as prior data
            state.add_labeling_data(record_ids=[paper_i],
                                    labels=[label],
                                    notes=[None],
                                    prior=prior)

        elif label == -1:
            with open_state(state_path, read_only=False) as state:
                state.delete_record_labeling_data(paper_i)

    if retrain_model:
        train_model(project_id)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import re
import shutil
import subprocess
import tempfile
import urllib.parse
import uuid
from collections import Counter
from pathlib import Path
from urllib.request import urlretrieve

from flask import Blueprint
from flask import Response
from flask import jsonify
from flask import request
from flask import send_file
from flask_cors import CORS
import numpy as np
import pandas as pd
from werkzeug.exceptions import InternalServerError
from werkzeug.utils import secure_filename

from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_FEATURE_EXTRACTION
from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import PROJECT_MODE_EXPLORE
from asreview.config import PROJECT_MODE_SIMULATE
from asreview.data import ASReviewData
from asreview.datasets import DatasetManager
from asreview.exceptions import BadFileFormatError
from asreview.models.balance import list_balance_strategies
from asreview.models.classifiers import list_classifiers
from asreview.models.feature_extraction import list_feature_extraction
from asreview.models.query import list_query_strategies
from asreview.search import SearchError
from asreview.search import fuzzy_find
from asreview.settings import ASReviewSettings
from asreview.state.errors import StateNotFoundError
from asreview.state.paths import get_data_file_path
from asreview.state.paths import get_data_path
from asreview.state.paths import get_lock_path
from asreview.state.paths import get_project_file_path
from asreview.state.paths import get_simulation_ready_path
from asreview.state.paths import get_state_path
from asreview.state.paths import get_tmp_path
from asreview.state.sql_converter import upgrade_asreview_project_file
from asreview.state.utils import open_state
from asreview.webapp.sqlock import SQLiteLock
from asreview.webapp.types import is_project
from asreview.webapp.utils.datasets import get_data_statistics
from asreview.webapp.utils.datasets import get_dataset_metadata
from asreview.webapp.utils.io import read_data
from asreview.webapp.utils.project import ProjectNotFoundError
from asreview.webapp.utils.project import _create_project_id
from asreview.webapp.utils.project import _get_executable
from asreview.webapp.utils.project import add_dataset_to_project
from asreview.webapp.utils.project import add_review_to_project
from asreview.webapp.utils.project import export_to_string
from asreview.webapp.utils.project import get_instance
from asreview.webapp.utils.project import get_paper_data
from asreview.webapp.utils.project import get_project_config
from asreview.webapp.utils.project import get_statistics
from asreview.webapp.utils.project import import_project_file
from asreview.webapp.utils.project import init_project
from asreview.webapp.utils.project import label_instance
from asreview.webapp.utils.project import remove_dataset_from_project
from asreview.webapp.utils.project import rename_project
from asreview.webapp.utils.project import update_instance
from asreview.webapp.utils.project import update_project_info
from asreview.webapp.utils.project import update_review_in_project
from asreview.webapp.utils.project_path import get_project_path
from asreview.webapp.utils.project_path import list_asreview_project_paths
from asreview.webapp.utils.validation import check_dataset

bp = Blueprint('api', __name__, url_prefix='/api')
CORS(bp, resources={r"*": {"origins": "*"}})

# error handlers


@bp.errorhandler(ProjectNotFoundError)
def project_not_found(e):

    message = str(e) if str(e) else "Project not found."
    logging.error(message)
    return jsonify(message=message), 400


@bp.errorhandler(InternalServerError)
def error_500(e):
    original = getattr(e, "original_exception", None)

    if original is None or str(e.original_exception) == "":
        # direct 500 error, such as abort(500)
        logging.error(e)
        return jsonify(message="Whoops, something went wrong."), 500

    # wrapped unhandled error
    logging.error(e.original_exception)
    return jsonify(message=str(e.original_exception)), 500


# routes
@bp.route('/projects', methods=["GET"])
def api_get_projects():  # noqa: F401
    """Get info on the article"""

    projects = list_asreview_project_paths()

    project_info = []
    for proj in projects:

        try:
            with open(proj / "project.json", "r") as f:
                res = json.load(f)

            # backwards support <0.10
            if "projectInitReady" not in res:
                res["projectInitReady"] = True

            # if time is not available (<0.14)
            if "created_at_unix" not in res:
                res["created_at_unix"] = None

            logging.info("Project found: {}".format(res["id"]))
            project_info.append(res)

        except Exception as err:
            logging.error(err)

    # sort the projects based on created_at_unix
    project_info = sorted(
        project_info,
        key=lambda y: (y["created_at_unix"] is not None, y["created_at_unix"]),
        reverse=True)

    response = jsonify({"result": project_info})
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/projects/stats', methods=["GET"])
def api_get_projects_stats():  # noqa: F401
    """Get dashboard statistics of all projects"""

    projects = list_asreview_project_paths()

    stats_counter = Counter()

    for proj in projects:

        try:
            with open(proj / "project.json", "r") as f:
                res = json.load(f)

            # backwards support <0.10
            if "projectInitReady" not in res:
                res["projectInitReady"] = True

            # get dashboard statistics
            statistics = get_statistics(res["id"])
            statistics["n_reviewed"] = statistics["n_included"] \
                + statistics["n_excluded"]

            if res["projectInitReady"] is not True:
                statistics["n_setup"] = 1
                statistics["n_in_review"] = 0
                statistics["n_finished"] = 0
            elif "reviewFinished" not in res or res[
                    "reviewFinished"] is not True:
                statistics["n_setup"] = 0
                statistics["n_in_review"] = 1
                statistics["n_finished"] = 0
            else:
                statistics["n_setup"] = 0
                statistics["n_in_review"] = 0
                statistics["n_finished"] = 1

            statistics = {
                x: statistics[x]
                for x in ("n_reviewed", "n_included", "n_setup", "n_in_review",
                          "n_finished")
            }
            stats_counter.update(statistics)

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to load dashboard statistics."), 500

    project_stats = dict(stats_counter)

    response = jsonify({"result": project_stats})
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/project/info', methods=["POST"])
def api_init_project():  # noqa: F401
    """Get info on the article"""

    project_mode = request.form['mode']
    project_name = request.form['name']
    project_description = request.form['description']
    project_authors = request.form['authors']

    project_id = _create_project_id(project_name)

    project_config = init_project(project_id,
                                  project_mode=project_mode,
                                  project_name=project_name,
                                  project_description=project_description,
                                  project_authors=project_authors)

    response = jsonify(project_config)

    return response, 201


@bp.route('/project/<project_id>/convert_if_old', methods=["GET"])
def api_convert_project_if_old(project_id):
    """Get if project is converted"""

    project_path = get_project_path(project_id)

    try:
        upgrade_asreview_project_file(project_path)

    except ValueError:
        pass

    except Exception as err:
        logging.error(err)
        message = "Failed to open the project in this version of ASReview LAB."
        return jsonify(message=message), 500

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/info', methods=["GET"])
def api_get_project_info(project_id):  # noqa: F401
    """Get info on the article"""
    project_path = get_project_path(project_id)
    project_info = get_project_config(project_id)

    try:

        # check if there is a dataset
        try:
            get_data_file_path(project_path)
            project_info["projectHasDataset"] = True
        except Exception:
            project_info["projectHasDataset"] = False

        # backwards support <0.10
        if "projectInitReady" not in project_info:
            if "projectHasPriorKnowledge" not in project_info:
                pass
            else:
                if project_info["projectHasPriorKnowledge"]:
                    project_info["projectInitReady"] = True
                else:
                    project_info["projectInitReady"] = False

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to retrieve project information."), 500

    return jsonify(project_info)


@bp.route('/project/<project_id>/info', methods=["PUT"])
def api_update_project_info(project_id):  # noqa: F401
    """Get info on the article"""

    # rename the project if project name is changed
    if request.form.get('name', None) is not None:
        project_id_new = rename_project(project_id, request.form['name'])

    # update the project info
    update_project_info(
        project_id_new,
        mode=request.form['mode'],
        description=request.form['description'],
        authors=request.form['authors'])

    return api_get_project_info(project_id_new)


@bp.route('/datasets', methods=["GET"])
def api_demo_data_project():  # noqa: F401
    """Get info on the article"""

    subset = request.args.get('subset', None)

    if subset == "plugin":

        try:
            result_datasets = get_dataset_metadata(
                exclude=["builtin", "benchmark"])

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to load plugin datasets."), 500

    elif subset == "benchmark":

        try:
            # collect the datasets metadata
            result_datasets = get_dataset_metadata(include="benchmark")

            # mark the featured datasets
            featured_dataset_ids = [
                "van_de_Schoot_2017", "Hall_2012", "Cohen_2006_ACEInhibitors",
                "Kwok_2020"
            ]
            for featured_id in featured_dataset_ids:
                for i, dataset in enumerate(result_datasets):
                    if result_datasets[i][
                            "dataset_id"] == f"benchmark:{featured_id}":
                        result_datasets[i]["featured"] = True

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to load benchmark datasets."), 500

    else:
        response = jsonify(message="demo-data-loading-failed")

        return response, 400

    payload = {"result": result_datasets}
    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/data', methods=["POST", "PUT"])
def api_upload_data_to_project(project_id):  # noqa: F401
    """Get info on the article"""
    project_path = get_project_path(project_id)

    # get the project config to modify behavior of dataset
    project_config = get_project_config(project_id)

    # remove old dataset if present
    if "dataset_path" in project_config and \
            project_config["dataset_path"] is not None:
        logging.warning("Removing old dataset and adding new dataset.")
        remove_dataset_from_project(project_id)

    # create dataset folder if not present
    get_data_path(project_path).mkdir(exist_ok=True)

    if request.form.get('plugin', None):
        url = DatasetManager().find(request.form['plugin']).url

    if request.form.get('benchmark', None):
        url = DatasetManager().find(request.form['benchmark']).url

    if request.form.get('url', None):
        url = request.form['url']

    if request.form.get('plugin', None) or request.form.get(
            'benchmark', None) or request.form.get('url', None):
        try:
            url_parts = urllib.parse.urlparse(url)
            filename = secure_filename(url_parts.path.rsplit('/', 1)[-1])

            urlretrieve(url, get_data_path(project_path) / filename)

        except ValueError as err:

            logging.error(err)
            message = f"Invalid URL '{url}'."

            if isinstance(url, str) and not url.startswith("http"):
                message += " Usually, the URL starts with 'http' or 'https'."

            return jsonify(message=message), 400

        except Exception as err:

            logging.error(err)
            message = f"Can't retrieve data from URL {url}."

            return jsonify(message=message), 400

    elif 'file' in request.files:

        data_file = request.files['file']

        # check the file is file is in a correct format
        check_dataset(data_file)
        try:

            filename = secure_filename(data_file.filename)
            fp_data = get_data_path(project_path) / filename

            # save the file
            data_file.save(str(fp_data))

        except Exception as err:

            logging.error(err)

            response = jsonify(
                message=f"Failed to upload file '{filename}'. {err}")

            return response, 400
    else:
        response = jsonify(message="No file or dataset found to upload.")
        return response, 400

    if project_config["mode"] == PROJECT_MODE_EXPLORE:

        data_path_raw = get_data_path(project_path) / filename
        data_path = data_path_raw.with_suffix('.csv')

        data = ASReviewData.from_file(data_path_raw)
        data.df.rename({data.column_spec["included"]: "debug_label"},
                       axis=1,
                       inplace=True)
        data.to_csv(data_path)
    else:
        data_path = get_data_path(project_path) / filename

    try:
        # add the file to the project
        add_dataset_to_project(project_id, data_path.name)

    # Bad format. TODO{Jonathan} Return informative message with link.
    except BadFileFormatError as err:
        message = f"Failed to upload file '{filename}'. {err}"
        return jsonify(message=message), 400

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/project/<project_id>/data', methods=["GET"])
def api_get_project_data(project_id):  # noqa: F401
    """Get info on the article"""
    project_path = get_project_path(project_id)

    if not is_project(project_id):
        response = jsonify(message="Project not found.")
        return response, 404

    try:

        filename = get_data_file_path(project_path).stem

        # get statistics of the dataset
        statistics = get_data_statistics(project_id)
        statistics["filename"] = filename

    except FileNotFoundError as err:
        logging.info(err)
        statistics = {"filename": None}

    except Exception as err:
        logging.error(err)
        message = f"Failed to get file. {err}"
        return jsonify(message=message), 400

    response = jsonify(statistics)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/search', methods=["GET"])
def api_search_data(project_id):  # noqa: F401
    """Search for papers
    """
    q = request.args.get('q', default=None, type=str)
    max_results = request.args.get('n_max', default=10, type=int)

    try:
        payload = {"result": []}
        if q:

            # read the dataset
            as_data = read_data(project_id)

            # search for the keywords
            result_idx = fuzzy_find(as_data,
                                    q,
                                    max_return=max_results,
                                    exclude=[],
                                    by_index=True)

            for paper in as_data.record(result_idx, by_index=True):
                payload["result"].append({
                    "id": int(paper.record_id),
                    "title": paper.title,
                    "abstract": paper.abstract,
                    "authors": paper.authors,
                    "keywords": paper.keywords,
                    "included": int(paper.included)
                })

    except SearchError as search_err:
        logging.error(search_err)
        return jsonify(message=f"Error: {search_err}"), 500

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load search results."), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/labeled', methods=["GET"])
def api_get_labeled(project_id):  # noqa: F401
    """Get all papers classified as labeled documents
    """

    page = request.args.get("page", default=None, type=int)
    per_page = request.args.get("per_page", default=20, type=int)
    subset = request.args.get("subset", default=None, type=str)
    latest_first = request.args.get("latest_first", default=1, type=int)

    project_path = get_project_path(project_id)

    try:

        with open_state(project_path) as s:
            data = s.get_dataset(["record_id", "label", "query_strategy"])
            data["prior"] = (data["query_strategy"] == "prior").astype(int)

        if subset in ["relevant", "included"]:
            data = data[data['label'] == 1]
        elif subset in ["irrelevant", "excluded"]:
            data = data[data['label'] == 0]
        else:
            data = data[~data['label'].isnull()]

        if latest_first == 1:
            data = data.iloc[::-1]

        # count labeled records and max pages
        count = len(data)
        max_page_calc = divmod(count, per_page)
        if max_page_calc[1] == 0:
            max_page = max_page_calc[0]
        else:
            max_page = max_page_calc[0] + 1

        if page is not None:
            # slice out records on specific page
            if page <= max_page:
                idx_start = page * per_page - per_page
                idx_end = page * per_page
                data = data.iloc[idx_start:idx_end, :].copy()
            else:
                raise ValueError(f"Page {page - 1} is the last page.")

            # set next & previous page
            if page < max_page:
                next_page = page + 1
                if page > 1:
                    previous_page = page - 1
                else:
                    previous_page = None
            else:
                next_page = None
                previous_page = page - 1
        else:
            next_page = None
            previous_page = None

        records = read_data(project_id).record(data["record_id"], by_index=False)

        payload = {
            "count": count,
            "next_page": next_page,
            "previous_page": previous_page,
            "result": [],
        }
        for i, record in zip(data.index.tolist(), records):

            payload["result"].append({
                "id": int(record.record_id),
                "title": record.title,
                "abstract": record.abstract,
                "authors": record.authors,
                "keywords": record.keywords,
                "included": int(data.loc[i, "label"]),
                "prior": int(data.loc[i, "prior"])
            })

    except Exception as err:
        logging.error(err)
        return jsonify(message=f"Failed to load labeled documents. {err}"), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/labeled_stats', methods=["GET"])
def api_get_labeled_stats(project_id):  # noqa: F401
    """Get all papers classified as prior documents
    """
    project_path = get_project_path(project_id)
    try:

        with open_state(project_path) as s:
            data = s.get_dataset(["label", "query_strategy"])
            data_prior = data[data["query_strategy"] == "prior"]

        response = jsonify({
            "n": len(data),
            "n_inclusions": sum(data['label'] == 1),
            "n_exclusions": sum(data['label'] == 0),
            "n_prior": len(data_prior),
            "n_prior_inclusions": sum(data_prior['label'] == 1),
            "n_prior_exclusions": sum(data_prior['label'] == 0)
        })
    except StateNotFoundError:
        response = jsonify({
            "n": 0,
            "n_inclusions": 0,
            "n_exclusions": 0,
            "n_prior": 0,
            "n_prior_inclusions": 0,
            "n_prior_exclusions": 0
        })

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load prior information."), 500

    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/prior_random', methods=["GET"])
def api_random_prior_papers(project_id):  # noqa: F401
    """Get a selection of random papers to find exclusions.

    This set of papers is extracted from the pool, but without
    the already labeled items.
    """
    project_path = get_project_path(project_id)
    state_file = get_state_path(project_path)

    with open_state(state_file) as state:
        pool, _, _ = state.get_pool_labeled_pending()

    try:
        pool_random = np.random.choice(pool, 1, replace=False)[0]
    except Exception:
        raise ValueError("Not enough random indices to sample from.")

    try:
        record = read_data(project_id).record(pool_random, by_index=False)

        payload = {"result": []}

        payload["result"].append({
            "id": int(record.record_id),
            "title": record.title,
            "abstract": record.abstract,
            "authors": record.authors,
            "keywords": record.keywords,
            "included": None
        })

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load random documents."), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/algorithms', methods=["GET"])
def api_list_algorithms():
    """List the names and labels of available algorithms"""

    try:
        classes = [
            list_balance_strategies(),
            list_classifiers(),
            list_feature_extraction(),
            list_query_strategies()
        ]

        payload = {
            "balance_strategy": [],
            "classifier": [],
            "feature_extraction": [],
            "query_strategy": [],
        }

        for c, key in zip(classes, payload.keys()):
            for method in c:
                if hasattr(method, "label"):
                    payload[key].append({
                        "name": method.name,
                        "label": method.label
                    })
                else:
                    payload[key].append({
                        "name": method.name,
                        "label": method.name
                    })

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to retrieve algorithms."), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/algorithms', methods=["GET"])
def api_get_algorithms(project_id):  # noqa: F401

    default_payload = {
        "model": DEFAULT_MODEL,
        "feature_extraction": DEFAULT_FEATURE_EXTRACTION,
        "query_strategy": DEFAULT_QUERY_STRATEGY
    }

    # check if there were algorithms stored in the state file
    try:

        project_path = get_project_path(project_id)
        state_file = get_state_path(project_path)

        with open_state(state_file) as state:
            if state.settings is not None:
                payload = {
                    "model": state.settings.model,
                    "feature_extraction": state.settings.feature_extraction,
                    "query_strategy": state.settings.query_strategy
                }
            else:
                payload = default_payload
    except (StateNotFoundError):
        payload = default_payload

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/algorithms', methods=["POST"])
def api_set_algorithms(project_id):  # noqa: F401

    # TODO@{Jonathan} validate model choice on server side
    ml_model = request.form.get("model", None)
    ml_query_strategy = request.form.get("query_strategy", None)
    ml_feature_extraction = request.form.get("feature_extraction", None)

    # create a new settings object from arguments
    # only used if state file is not present
    asreview_settings = ASReviewSettings(
        mode="minimal",
        model=ml_model,
        query_strategy=ml_query_strategy,
        balance_strategy=DEFAULT_BALANCE_STRATEGY,
        feature_extraction=ml_feature_extraction)

    project_path = get_project_path(project_id)
    state_file = get_state_path(project_path)

    # save the new settings to the state file
    with open_state(state_file, read_only=False) as state:
        state.settings = asreview_settings

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/start', methods=["POST"])
def api_start(project_id):  # noqa: F401
    """Start training of first model or simulation.
    """
    project_path = get_project_path(project_id)
    project_config = get_project_config(project_id)

    # the project is a simulation project
    if project_config["mode"] == PROJECT_MODE_SIMULATE:

        logging.info("Starting simulation")

        try:
            simulation_id = uuid.uuid4().hex
            datafile = get_data_file_path(project_path)
            state_file = get_simulation_ready_path(project_path, simulation_id)

            logging.info("Project data file found: {}".format(datafile))

            add_review_to_project(project_id, simulation_id)

            # start simulation
            py_exe = _get_executable()
            run_command = [
                py_exe, "-m", "asreview", "simulate", datafile, "--state_file",
                state_file
            ]
            subprocess.Popen(run_command)

        except Exception as err:
            logging.error(err)
            message = f"Failed to get data file. {err}"
            return jsonify(message=message), 400

    # the project is an oracle or explore project
    else:

        logging.info("Train first iteration of model")
        try:
            # start training the model
            py_exe = _get_executable()
            run_command = [
                # get executable
                py_exe,
                # get module
                "-m", "asreview",
                # train the model via cli
                "web_run_model",
                # specify project id
                project_id,
                # output the error of the first model
                "--output_error"
            ]
            subprocess.Popen(run_command)

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to train the model."), 500

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/ready', methods=["GET"])
def api_init_model_ready(project_id):  # noqa: F401
    """Check if trained model is available
    """
    project_path = get_project_path(project_id)

    project_config = get_project_config(project_id)

    if project_config["mode"] == PROJECT_MODE_SIMULATE:
        logging.info("checking if simulation is ready")

        simulation_id = project_config["reviews"][0]["id"]

        if get_simulation_ready_path(project_path, simulation_id).exists():
            logging.info("simulation ready")
            update_review_in_project(project_id, simulation_id, "ready")

            response = jsonify({'status': 1})
        else:
            logging.info("simulation not ready")
            response = jsonify({'status': 0})

    else:
        error_path = get_project_path(project_id) / "error.json"
        if error_path.exists():
            logging.error("error on training")
            with open(error_path, "r") as f:
                error_message = json.load(f)
            return jsonify(message=error_message), 400

        try:
            with open_state(project_path) as state:
                if state.model_has_trained:
                    update_project_info(project_id, projectInitReady=True)

                    response = jsonify({'status': 1})
                else:
                    response = jsonify({'status': 0})

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to initiate the project."), 500

    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


# TODO{Terry}: This may be deprecated when the new state file is in use.
# @bp.route('/project/<project_id>/model/clear_error', methods=["DELETE"])
# def api_clear_model_error(project_id):
#     """Clear model training error"""

#     error_path = get_project_path(project_id) / "error.json"
#     project_path = get_project_path(project_id)
#     state_path = get_state_path(project_path)

#     if error_path.exists() and state_path.exists():
#         os.remove(error_path)
#         os.remove(state_path)

#         response = jsonify({'success': True})
#         response.headers.add('Access-Control-Allow-Origin', '*')
#         return response

#     response = jsonify(message="Failed to clear model training error.")
#     return response, 500


@bp.route('/project/import_project', methods=["POST"])
def api_import_project():
    """Import uploaded project"""

    # raise error if file not given
    if 'file' not in request.files:
        response = jsonify(message="No file found to upload.")
        return response, 400

    # set the project file
    project_file = request.files['file']
    # import the project
    project_id = import_project_file(project_file)

    # return the project info in the same format as project_info
    return jsonify(id=project_id)


@bp.route('/project/<project_id>/export', methods=["GET"])
def export_results(project_id):
    project_path = get_project_path(project_id)

    # get the export args
    file_type = request.args.get('file_type', None)

    try:
        # CSV
        if file_type == "csv":
            dataset_str = export_to_string(project_id, export_type="csv")

            return Response(
                dataset_str,
                mimetype="text/csv",
                headers={
                    "Content-disposition":
                    f"attachment; filename=asreview_result_{project_id}.csv"
                })
        # TSV
        elif file_type == "tsv":
            dataset_str = export_to_string(project_id, export_type="tsv")

            return Response(
                dataset_str,
                mimetype="text/tab-separated-values",
                headers={
                    "Content-disposition":
                    f"attachment; filename=asreview_result_{project_id}.tsv"
                })
        # Excel
        elif file_type == "xlsx":
            dataset_str = export_to_string(project_id, export_type="excel")
            fp_tmp_export = Path(get_tmp_path(project_path), "export_result.xlsx")

            return send_file(
                fp_tmp_export,
                mimetype=   # noqa
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",  # noqa
                as_attachment=True,
                download_name=f"asreview_result_{project_id}.xlsx",
                max_age=0)
        # RIS
        elif file_type == "ris":
            if get_data_file_path(project_id).suffix not in [
                    ".ris", ".RIS", ".txt", ".TXT"
            ]:
                raise ValueError(
                    "RIS file can be exported only when RIS file was imported.")

            dataset_str = export_to_string(project_id, export_type="ris")

            return Response(
                dataset_str,
                mimetype="application/octet-stream",
                headers={
                    "Content-disposition":
                    f"attachment; filename=asreview_result_{project_id}.ris"
                })

        else:
            raise TypeError("File type should be: .ris/.csv/.tsv/.xlsx")
    except Exception as err:
        logging.error(err)
        return jsonify(message=f"Failed to export the {file_type} dataset."), 500


@bp.route('/project/<project_id>/export_project', methods=["GET"])
def export_project(project_id):
    """Export the project file.

    The ASReview project file is a file with .asreview extension.
    The ASReview project file is a zipped file and contains
    all information to continue working on the project as well
    as the orginal dataset.
    """

    # create a temp folder to zip
    tmpdir = tempfile.TemporaryDirectory()

    # copy the source tree, but ignore pickle files
    shutil.copytree(get_project_path(project_id),
                    Path(tmpdir.name, project_id),
                    ignore=shutil.ignore_patterns('*.pickle'))

    # create the archive
    shutil.make_archive(Path(tmpdir.name, project_id),
                        "zip",
                        root_dir=Path(tmpdir.name, project_id))

    try:
        # return the project file to the user
        return send_file(str(Path(tmpdir.name, f"{project_id}.zip")),
                         as_attachment=True,
                         download_name=f"{project_id}.asreview",
                         max_age=0)

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to export the project."), 500


@bp.route('/project/<project_id>/finish', methods=["GET"])
def api_finish_project(project_id):
    """Mark a project as finished or not"""
    project_path = get_project_path(project_id)

    # read the file with project info
    with open(get_project_file_path(project_path), "r") as fp:
        project_info = json.load(fp)

    try:
        project_info["reviewFinished"] = not project_info["reviewFinished"]
    except KeyError:
        # missing key in projects created in older versions
        project_info["reviewFinished"] = True

    # update the file with project info
    with open(get_project_file_path(project_path), "w") as fp:
        json.dump(project_info, fp)

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


# @bp.route('/project/<project_id>/document/<doc_id>/info', methods=["GET"])
# def api_get_article_info(project_id, doc_id=None):  # noqa: F401
#     """Get info on the article"""

#     data = get_paper_data(
#         project_id,
#         paper_id=doc_id,
#         return_debug_label=True
#     )

#     response = jsonify(data)
#     response.headers.add('Access-Control-Allow-Origin', '*')

#     return response


@bp.route('/project/<project_id>/progress', methods=["GET"])
def api_get_progress_info(project_id):  # noqa: F401
    """Get progress statistics of a project"""

    try:
        statistics = get_statistics(project_id)

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load progress statistics."), 500

    response = jsonify(statistics)
    response.headers.add('Access-Control-Allow-Origin', '*')

    # return a success response to the client.
    return response


@bp.route('/project/<project_id>/progress_density', methods=["GET"])
def api_get_progress_density(project_id):
    """Get progress density of a project"""

    try:
        # get label history
        project_path = get_project_path(project_id)

        with open_state(project_path) as s:
            data = s.get_labels()

        # create a dataset with the rolling mean of every 10 papers
        df = data \
            .to_frame(name="Relevant") \
            .rolling(10, min_periods=1) \
            .mean()
        df["Total"] = df.index + 1

        # transform mean(percentage) to number
        for i in range(0, len(df)):
            if df.loc[i, "Total"] < 10:
                df.loc[i, "Irrelevant"] = (
                    1 - df.loc[i, "Relevant"]) * df.loc[i, "Total"]
                df.loc[i,
                       "Relevant"] = df.loc[i, "Total"] - df.loc[i,
                                                                 "Irrelevant"]
            else:
                df.loc[i, "Irrelevant"] = (1 - df.loc[i, "Relevant"]) * 10
                df.loc[i, "Relevant"] = 10 - df.loc[i, "Irrelevant"]

        df = df.round(1).to_dict(orient="records")
        for d in df:
            d["x"] = d.pop("Total")

        df_relevant = [{k: v
                        for k, v in d.items() if k != "Irrelevant"}
                       for d in df]
        for d in df_relevant:
            d["y"] = d.pop("Relevant")

        df_irrelevant = [{k: v
                          for k, v in d.items() if k != "Relevant"}
                         for d in df]
        for d in df_irrelevant:
            d["y"] = d.pop("Irrelevant")

        payload = {"relevant": df_relevant, "irrelevant": df_irrelevant}

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load progress density."), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/project/<project_id>/progress_recall', methods=["GET"])
def api_get_progress_recall(project_id):
    """Get cumulative number of inclusions by ASReview/at random"""

    project_path = get_project_path(project_id)
    try:
        with open_state(project_path) as s:
            data = s.get_labels()
            n_records = len(s.get_record_table())

        # create a dataset with the cumulative number of inclusions
        df = data \
            .to_frame(name="Relevant") \
            .cumsum()
        df["Total"] = df.index + 1
        df["Random"] = (df["Total"] *
                        (df["Relevant"][-1:] / n_records).values).round()

        df = df.round(1).to_dict(orient="records")
        for d in df:
            d["x"] = d.pop("Total")

        df_asreview = [{k: v
                        for k, v in d.items() if k != "Random"} for d in df]
        for d in df_asreview:
            d["y"] = d.pop("Relevant")

        df_random = [{k: v
                      for k, v in d.items() if k != "Relevant"} for d in df]
        for d in df_random:
            d["y"] = d.pop("Random")

        payload = {"asreview": df_asreview, "random": df_random}

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to load progress recall."), 500

    response = jsonify(payload)
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/project/<project_id>/record/<doc_id>', methods=["POST", "PUT"])
def api_classify_instance(project_id, doc_id):  # noqa: F401
    """Label item

    This request handles the document identifier and the corresponding label.
    The result is stored in a temp location. If this storage exceeds a certain
    amount of values, then the model is triggered. The values of the location
    are passed to the model and the storaged is cleared. This model will run
    in the background.
    """
    # return the combination of document_id and label.
    doc_id = request.form.get('doc_id')
    label = request.form.get('label')
    is_prior = request.form.get('is_prior', default=False)

    retrain_model = False if is_prior == "1" else True
    prior = True if is_prior == "1" else False

    if request.method == 'POST':
        label_instance(project_id,
                       doc_id,
                       label,
                       prior=prior,
                       retrain_model=retrain_model)
    elif request.method == 'PUT':
        update_instance(project_id,
                        doc_id,
                        label,
                        retrain_model=retrain_model)

    response = jsonify({'success': True})
    response.headers.add('Access-Control-Allow-Origin', '*')

    return response


@bp.route('/project/<project_id>/get_document', methods=["GET"])
def api_get_document(project_id):  # noqa: F401
    """Retrieve documents in order of review.

    After these documents were retrieved, the queue on the client side is
    updated.
    This request can get triggered after each document classification.
    Although it might be better to call this function after 20 requests on the
    client side.
    """
    try:
        new_instance = get_instance(project_id)

        if new_instance is None:  # don't use 'if not new_instance:'

            item = None
            pool_empty = True
        else:

            item = get_paper_data(project_id,
                                  new_instance,
                                  return_debug_label=True)
            item["doc_id"] = new_instance
            pool_empty = False

    except Exception as err:
        logging.error(err)
        return jsonify(message="Failed to retrieve new documents."), 500

    response = jsonify({"result": item, "pool_empty": pool_empty})
    response.headers.add('Access-Control-Allow-Origin', '*')
    return response


@bp.route('/project/<project_id>/delete', methods=["DELETE"])
def api_delete_project(project_id):  # noqa: F401
    """Get info on the article"""

    # some checks to check if there is a project to delete
    if project_id == "" or project_id is None:
        response = jsonify(message="project-delete-failure")
        return response, 500

    project_path = get_project_path(project_id)

    if project_path.exists() and project_path.is_dir():
        shutil.rmtree(project_path)

        response = jsonify({'success': True})
        response.headers.add('Access-Control-Allow-Origin', '*')
        return response

    response = jsonify(message="project-delete-failure")
    return response, 500



# Copyright 2019-2021 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import pandas as pd
import rispy


def write_ris(df, fp):
    """RIS file writer.

    Parameters
    ----------
    df: pandas.Dataframe
        Dataframe to convert and export.
    fp: str, pathlib.Path
        File path to the RIS file, if exists.

    Returns
    -------
    rispy.dumps:
        Dataframe with entries is written into buffer.
    rispy.dump:
        Dataframe with entries is written into a file.
    """

    # Turn pandas DataFrame into records (list of dictionaries) for rispy
    records = df.to_dict('records')

    # Create an array for storing modified records
    records_new = []

    # Iterate over all available records
    for rec in records:

        # Store the record as a deepcopy
        rec_copy = copy.deepcopy(rec)

        # Iterate over all the items for the deepcopied record
        for k, v in rec_copy.items():
            # Find all items with a value
            if not isinstance(v, list) and pd.isnull(v):
                # Assign the value to the key
                rec_copy[k] = v

        # Check the "authors" - AU
        try:
            rec_copy["authors"] = eval(rec_copy["authors"])
        except Exception:
            rec_copy["authors"] = []

        # Check the "keywords" - KW
        try:
            rec_copy["keywords"] = eval(rec_copy["keywords"])
        except Exception:
            rec_copy["keywords"] = []

        # Check the "notes" - N1
        try:
            rec_copy["notes"] = eval(str(rec_copy["notes"]))
        except Exception:
            rec_copy["notes"] = []

        # Update "notes" column based on the "included" column label
        finally:
            # Relevant records
            if "included" in rec_copy and rec_copy["included"] == 1:
                rec_copy["notes"].append("ASReview_relevant")
            # Irrelevant records
            elif "included" in rec_copy and rec_copy["included"] == 0:
                rec_copy["notes"].append("ASReview_irrelevant")
            # Not seen records
            elif "included" in rec_copy and rec_copy["included"] == -1:
                rec_copy["notes"].append("ASReview_not_seen")
            else:
                rec_copy["notes"].append("ASReview_not_seen")

        # Append the deepcopied and updated record to a new array
        records_new.append(rec_copy)

    # From buffered dataframe
    if fp is None:
        # Write the whole content to buffer
        return rispy.dumps(records_new)

    # From IO dataframe
    else:
        # Write the whole content to a file
        with open(fp, "w", encoding="utf8") as fp:
            rispy.dump(records_new, fp)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

import pandas
import rispy

from asreview.io.utils import _standardize_dataframe


def _strip_zotero_p_tags(note_list):
    """Converter function for removing the XHTML <p></p> tags from Zotero export.

    Parameters
    ----------
    note_list: list
        A list of notes, coming from the Dataframe's "notes" column.

    Returns
    -------
    new_notes: list
        A list of updated notes, where XHTML <p></p> tags have been stripped.
    note_list: list
        The original note_list, when no XHTML <p></p> tags have been found.
    """
    if isinstance(note_list, list):
        new_notes = []
        for v in note_list:
            try:
                new_notes.append(re.sub(r'^<p>|<\/p>$', '', v))
            except Exception:
                new_notes.append(v)
        return new_notes
    else:
        return note_list


def _label_parser(note_list):
    """Converter function for manipulating the internal "included" and "notes" columns.

    Parameters
    ----------
    note_list: list
        A list of notes, coming from the Dataframe's "notes" column.

    Returns
    -------
    asreview_new_notes: list
        A list of updated notes, where internal label has been added.
    note_list: list
        The original note_list, when no labels have been found.
    1,0,-1: int
        Labels in case they are still needed from the internal representation.
    """
    regex = r"ASReview_relevant|ASReview_irrelevant|ASReview_not_seen"

    # Check whether note_list is actually a list and not NaN
    # Return -1 and an empty list
    if not isinstance(note_list, list):
        return -1, []

    # Create lists of lists for ASReview references
    asreview_refs = [re.findall(regex, note) for note in note_list]
    asreview_refs_list = [
        item for sublist in asreview_refs for item in sublist
    ]

    if len(asreview_refs_list) > 0:
        # Create lists of lists for notes without references
        asreview_new_notes = [re.sub(regex, "", note) for note in note_list]
        # Remove empty elements from list
        asreview_new_notes[:] = [
            item for item in asreview_new_notes if item != ""
        ]
        label = asreview_refs_list[-1]

        # Check for the label and return proper values for internal representation
        if label == "ASReview_relevant":
            return 1, asreview_new_notes
        elif label == "ASReview_irrelevant":
            return 0, asreview_new_notes
        elif label == "ASReview_not_seen":
            return -1, asreview_new_notes
    else:
        return -1, note_list


def read_ris(fp):
    """RIS file reader.

    Parameters
    ----------
    fp: str, pathlib.Path
        File path to the RIS file.

    Returns
    -------
    pandas.DataFrame:
        Dataframe with entries.

    Raises
    ------
    ValueError
        File with unrecognized encoding is used as input.
    """

    encodings = ['utf-8', 'utf-8-sig', 'ISO-8859-1']
    entries = None
    for encoding in encodings:
        try:
            with open(fp, 'r', encoding=encoding) as bibliography_file:
                entries = list(
                    rispy.load(bibliography_file, skip_unknown_tags=True))
                break
        except UnicodeDecodeError:
            pass
        except IOError as e:
            logging.warning(e)
    if entries is None:
        raise ValueError("Cannot find proper encoding for data file.")

    # Turn the entries dictionary into a Pandas dataframe
    df = pandas.DataFrame(entries)

    # Check if "notes" column is present
    if "notes" in df:
        # Strip Zotero XHTML <p> tags on "notes"
        df["notes"] = df["notes"].apply(_strip_zotero_p_tags)
        # Split "included" from "notes"
        df[["included", "notes"
            ]] = pandas.DataFrame(df["notes"].apply(_label_parser).tolist(),
                                  columns=["included", "notes"])
        # Return the standardised dataframe with label and notes separated
        return _standardize_dataframe(df)
    else:
        # Return the standardised dataframe
        return _standardize_dataframe(df)



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import shutil
import sqlite3
import time
from base64 import b64decode
from contextlib import contextmanager
from datetime import datetime
from io import BytesIO
from pathlib import Path
from uuid import uuid4

import pandas as pd
from scipy.sparse import csr_matrix
from scipy.sparse import load_npz

from asreview._version import get_versions
from asreview.state.errors import StateNotFoundError
from asreview.state.paths import get_data_path
from asreview.state.paths import get_feature_matrices_path
from asreview.state.paths import get_project_file_path
from asreview.state.paths import get_reviews_path
from asreview.state.sqlstate import SqlStateV1

V3STATE_VERSION = "1.0"


# TODO(State): Create an 'add_project_json' function.
def is_zipped_project_file(fp):
    """Check if it is a zipped asreview project file."""
    if Path(fp).is_file():
        state_ext = Path(fp).suffix

        # TODO(State): Make link.
        if state_ext in ['.h5', '.hdf5', '.he5', '.json']:
            raise ValueError(
                f'State file with extension {state_ext} is no longer '
                f'supported. Migrate to the new format or '
                'use an older version of ASReview. See LINK.')
        elif state_ext == '.asreview':
            return True
        else:
            raise ValueError(f'State file extension {state_ext} is not '
                             f'recognized.')
    else:
        return False


def is_valid_project_folder(fp):
    """Check of the folder contains an asreview project."""
    if not Path(fp, 'reviews').is_dir() \
            or not Path(fp, 'feature_matrices').is_dir():
        raise ValueError(f"There does not seem to be a valid project folder"
                         f" at {fp}. The 'reviews' or 'feature_matrices' "
                         f"folder is missing.")
    else:
        return


def init_project_folder_structure(project_path,
                                  project_mode="oracle",
                                  project_name=None,
                                  project_description=None,
                                  project_authors=None):
    """Initialize a project folder structure at the given filepath.

    Arguments
    ---------
    project_path: pathlike
        Filepath where to intialize the project folder structure.
    project_mode: str
        Mode of the project. Should be 'oracle', 'explore' or 'simulate'.
    project_name: str
    project_description: str
    project_authors: str

    Returns
    -------
    dict
        Project configuration dictionary.
    """
    project_path = Path(project_path)
    project_id = project_path.stem
    asreview_version = get_versions()['version']

    if not project_id and not isinstance(project_id, str) \
            and len(project_id) >= 3:
        raise ValueError("Project name should be at least 3 characters.")

    if project_path.is_dir():
        raise IsADirectoryError(
            f'Project folder {project_path} already exists.')

    try:
        project_path.mkdir(exist_ok=True)
        get_data_path(project_path).mkdir(exist_ok=True)
        get_feature_matrices_path(project_path).mkdir(exist_ok=True)
        get_reviews_path(project_path).mkdir(exist_ok=True)

        project_config = {
            'version': asreview_version,  # todo: Fail without git?
            'id': project_id,
            'mode': project_mode,
            'name': project_name,
            'description': project_description,
            'authors': project_authors,
            'created_at_unix': int(time.time()),

            # project related variables
            'datetimeCreated': str(datetime.now()),
            'projectInitReady': False,
            'reviewFinished': False,
            'reviews': [],
            'feature_matrices': []
        }

        # create a file with project info
        with open(get_project_file_path(project_path), "w") as project_path:
            json.dump(project_config, project_path)

        return project_config

    except Exception as err:
        # remove all generated folders and raise error
        shutil.rmtree(project_path)
        raise err


@contextmanager
def open_state(working_dir, review_id=None, read_only=True):
    """Initialize a state class instance from a project folder.

    Arguments
    ---------
    working_dir: str/pathlike
        Filepath to the (unzipped) project folder.
    review_id: str
        Identifier of the review from which the state will be instantiated.
        If none is given, the first review in the reviews folder will be taken.
    read_only: bool
        Whether to open in read_only mode.

    Returns
    -------
    SqlStateV1
    """
    working_dir = Path(working_dir)

    if not get_reviews_path(working_dir).is_dir():
        if read_only:
            raise StateNotFoundError(f"There is no valid project folder"
                                     f" at {working_dir}")
        else:
            init_project_folder_structure(working_dir)
            review_id = uuid4().hex

    # Check if file is a valid project folder.
    is_valid_project_folder(working_dir)

    # Get the review_id of the first review if none is given.
    # If there is no review yet, create a review id.
    if review_id is None:
        reviews = list(get_reviews_path(working_dir).iterdir())
        if reviews:
            review_id = reviews[0].name
        else:
            review_id = uuid4().hex

    # init state class
    state = SqlStateV1(read_only=read_only)

    try:
        if Path(get_reviews_path(working_dir), review_id).is_dir():
            state._restore(working_dir, review_id)
        elif not Path(get_reviews_path(working_dir), review_id).is_dir() \
                and not read_only:
            state._create_new_state_file(working_dir, review_id)
        else:
            raise StateNotFoundError("State file does not exist, and in "
                                     "read only mode.")
        yield state
    finally:
        try:
            state.close()
        except AttributeError:
            # file seems to be closed, do nothing
            pass


def delete_state_from_project(project_path, remove_folders=False):

    project_file_path = get_project_file_path(project_path)

    try:
        path_fm = get_feature_matrices_path(project_path)
        shutil.rmtree(path_fm)
        if not remove_folders:
            get_feature_matrices_path(project_path).mkdir(exist_ok=True)
    except Exception:
        print("Failed to remove feature matrices.")

    try:
        path_review = get_reviews_path(project_path)
        shutil.rmtree(path_review)
        if not remove_folders:
            get_reviews_path(project_path).mkdir(exist_ok=True)
    except Exception:
        print("Failed to remove sql database.")

    # open the projects file
    with open(project_file_path, "r") as f_read:
        project_dict = json.load(f_read)

    # add path to dict (overwrite if already exists)
    reset_config = {
        'projectInitReady': False,
        'reviewFinished': False,
        'reviews': [],
        'feature_matrices': []
    }
    project_dict.update(reset_config)

    with open(project_file_path, "w") as f_write:
        json.dump(project_dict, f_write)


def read_results_into_dataframe(fp, table='results'):
    """Read the result table of a v3 state file into a pandas dataframe.

    Arguments
    ---------
    fp: str
        Project folder.
    table: str
        Name of the sql table in the results.sql that you want to read.

    Returns
    -------
    pd.DataFrame
        Dataframe containing contents of the results table of the state file.
    """
    path = Path(fp)
    con = sqlite3.connect(path / 'results.sql')
    df = pd.read_sql_query(f'SELECT * FROM {table}', con)
    con.close()
    return df


def decode_feature_matrix(jsonstate, data_hash):
    """Get the feature matrix from a json state as a scipy csr_matrix."""
    my_data = jsonstate._state_dict["data_properties"][data_hash]
    encoded_X = my_data["feature_matrix"]
    matrix_type = my_data["matrix_type"]
    if matrix_type == "ndarray":
        return csr_matrix(encoded_X)
    elif matrix_type == "csr_matrix":
        with BytesIO(b64decode(encoded_X)) as f:
            return load_npz(f)
    return encoded_X



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path


def get_project_file_path(project_path):
    """Get the path to the project file.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return Path(project_path, "project.json")


def get_tmp_path(project_path):
    """Get the tmp directory in the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return Path(project_path, "tmp")


def get_data_path(project_path):
    """Get the path to the data folder.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return Path(project_path, "data")


def get_data_file_path(project_path):

    data_folder = get_data_path(project_path)
    project_file_path = get_project_file_path(project_path)

    try:
        # open the projects file
        with open(project_file_path, "r") as f_read:
            project_dict = json.load(f_read)
            data_filename = project_dict["dataset_path"]

    except Exception:
        raise Exception("Dataset location not found")

    return data_folder / data_filename


def get_lock_path(project_path):
    """Get the active file for the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """
    return Path(project_path, "lock.sqlite")


def get_pool_path(project_path):
    """Get the pool file for the project and iteration.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return Path(project_path, "pool.json")


def get_labeled_path(project_path):
    """Get the labeled file for the project and iteration.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return Path(project_path, "labeled.json")


def get_reviews_path(project_path):
    """Get the reviews folder from the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """
    return Path(project_path, 'reviews')


def get_feature_matrices_path(project_path):
    """Get the feature matrices folder from the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """
    return Path(project_path, 'feature_matrices')


def get_sql_path(project_path, review_id=None):
    """Get the results sql file from the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    review_id: str
        Identifier for the review from which to get the results sql file.
        If none is given, pick the first id available.
    """
    if review_id is None:
        with open(get_project_file_path(project_path), 'r') as f:
            project_config = json.load(f)
        review_id = project_config['reviews'][0]['id']

    return Path(get_reviews_path(project_path), review_id, 'results.sql')


def get_settings_metadata_path(project_path, review_id=None):
    """Get the settings/metadata json file from the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    review_id: str
        Identifier for the review from which to get the settings json file.
        If none is given, pick the first id available.
    """
    if review_id is None:
        with open(get_project_file_path(project_path), 'r') as f:
            project_config = json.load(f)
        review_id = project_config['reviews'][0]['id']

    return Path(get_reviews_path(project_path), review_id,
                'settings_metadata.json')


def get_feature_matrix_path(project_path, feature_extraction=None):
    """Get the feature matrix file from the project.

    Arguments
    ---------
    project_path: str
        The path to the project.
    feature_extraction: str
        Identifier for the feature extraction method. If none is given,
        the first available feature matrix is used.
    """
    with open(get_project_file_path(project_path), 'r') as f:
        project_config = json.load(f)

    # Get the filename for the matrix with the given feature extraction method.
    if feature_extraction is None:
        filename = project_config['feature_matrices'][0]['filename']
    else:
        filename = next(config['filename']
                        for config in project_config['feature_matrices']
                        if config['id'] == feature_extraction)

    return Path(get_feature_matrices_path(project_path), filename)


# TODO(State): Merge with get_project_path.
def get_state_path(project_path):
    """Get the labeled file for the project and iteration.

    Arguments
    ---------
    project_path: str
        The path to the project.
    """

    return project_path


def get_simulation_ready_path(project_path, simulation_id):
    """Get the simulation_ready file for the project and iteration.

    Arguments
    ---------
    project_path: str
        The path to the project.
    simulation_id: str
        The id of the current simulation.
    """

    return Path(project_path, "reviews", simulation_id + ".json")



# Copyright 2019-2020 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sqlite3
from pathlib import Path
from uuid import uuid4

import numpy as np
from scipy.sparse import issparse
from scipy.sparse import save_npz

from asreview.state.errors import StateError
from asreview.state.legacy.utils import open_state as open_state_legacy
from asreview.state.utils import decode_feature_matrix

SQLSTATE_VERSION = "1.0"
ASREVIEW_FILE_EXTENSION = '.asreview'


def is_old_project(fp):
    """Check if state file is old version."""
    if Path(fp, 'reviews').is_dir():
        raise ValueError(f"There already is a 'reviews' folder at {fp}. "
                         f"This project seems to be in new format.")
    if not Path(fp, 'result.json').is_file():
        raise ValueError(f"There is no 'result.json' file at {fp}")


# TODO(State): Allow basic/full (i.e. save probabilities).
def upgrade_asreview_project_file(fp, from_version=0, to_version=1):
    """Convert an old asreview project folder to the new format.

    Arguments
    ---------
    fp: str/path
        Location of the (unzipped) project file.

    Returns
    -------
    Converts the data in the project to the new format
    and adds it to the folder in place."""

    if from_version != 0 and to_version != 1:
        raise ValueError(
            f"Not possible to upgrade from {from_version} to {to_version}.")

    # Check if it is indeed an old format project.
    is_old_project(fp)

    # Current Paths
    fp = Path(fp)
    json_fp = Path(fp, 'result.json')
    project_fp = Path(fp, 'project.json')
    pool_fp = Path(fp, 'pool.json')
    kwargs_fp = Path(fp, 'kwargs.json')
    review_id = str(uuid4().hex)

    # Create the reviews folder and the paths for the results and settings.
    Path(fp, 'reviews', review_id).mkdir(parents=True)
    sql_fp = str(Path(fp, 'reviews', review_id, 'results.sql'))
    settings_metadata_fp = Path(fp, 'reviews', review_id,
                                'settings_metadata.json')

    # Create the path for the feature matrix.

    # Create sqlite table with the results of the review.
    convert_json_results_to_sql(sql_fp, json_fp)

    # Create sqlite tables 'last_probabilities'.
    convert_json_last_probabilities(sql_fp, json_fp)

    # Create teh table for the last ranking of the model.
    create_last_ranking_table(sql_fp, pool_fp, kwargs_fp)

    # Add the record table to the sqlite database as the table
    # 'record_table'.
    convert_json_record_table(sql_fp, json_fp)

    # Create decision changes table.
    create_decision_changes_table(sql_fp)

    # Create json for settings.
    convert_json_settings_metadata(settings_metadata_fp, json_fp)

    # Create file for the feature matrix.
    with open(Path(fp, 'kwargs.json'), 'r') as f:
        kwargs_dict = json.load(f)
        feature_extraction_method = kwargs_dict['feature_extraction']
    feature_matrix_fp = convert_json_feature_matrix(fp, json_fp,
                                                    feature_extraction_method)

    # Update the project.json file.
    with open(json_fp, 'r') as f:
        start_time = json.load(f)['time']['start_time']
    convert_project_json(project_fp, review_id, start_time, feature_matrix_fp,
                         feature_extraction_method)


def convert_project_json(project_fp, review_id, start_time, feature_matrix_fp,
                         feature_extraction_method):
    """Update the project.json file to contain the review information , the
    feature matrix information and the new state version number.

    Arguments
    ---------
    project_fp: str/path
        Path to the project json file.
    review_id: str
        Identifier of the review.
    start_time: str
        String containing start time of the review.
    feature_matrix_fp: str/path
        Location of the feature matrix.
    feature_extraction_method: str
        Name of the feature extraction method.
    """
    with open(project_fp, 'r') as f:
        project_info = json.load(f)

    # Add the feature matrix information.
    feature_matrix_name = Path(feature_matrix_fp).name
    project_info['feature_matrices'] = [{
        'id': feature_extraction_method,
        'filename': feature_matrix_name
    }]

    # Add the review information.
    project_info['reviews'] = [{
        'id':
        review_id,
        'start_time':
        start_time,
        'review_finished':
        project_info.get('reviewFinished', False)
    }]

    # Update the state version.
    project_info['state_version'] = SQLSTATE_VERSION

    with open(project_fp, 'w') as f:
        json.dump(project_info, f)


def convert_json_settings_metadata(fp, json_fp):
    """Get the settings and metadata from a json state and save it as
    a json file at the location given by fp.

    Arguments
    ---------
    fp: str/path
        Path where to save the json file.
    json_fp: str/path
        Path to the json state file.
    """
    data_dict = {}
    with open_state_legacy(json_fp) as json_state:
        data_dict['settings'] = json_state._state_dict['settings']
        data_dict['current_queries'] = json_state._state_dict[
            'current_queries']
        data_dict['state_version'] = SQLSTATE_VERSION
        data_dict['software_version'] = json_state._state_dict[
            'software_version']
    with open(fp, 'w') as f:
        json.dump(data_dict, f)


def create_last_ranking_table(sql_fp, pool_fp, kwargs_fp):
    """Create the table which will contain the ranking of the last iteration of
    the model. The converter will leave the table empty. It will be filled the
    first time a new model is trained.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    """

    with open(pool_fp) as f_pool:
        pool_ranking = json.load(f_pool)

    with open(kwargs_fp, 'r') as f_kwargs:
        kwargs_dict = json.load(f_kwargs)

    last_ranking = [(v, i, kwargs_dict['model'], kwargs_dict['query_strategy'],
                     kwargs_dict['balance_strategy'],
                     kwargs_dict['feature_extraction'], None, None)
                    for i, v in enumerate(pool_ranking)]

    with sqlite3.connect(sql_fp) as con:
        cur = con.cursor()

        # Create the last_ranking table.
        cur.execute('''CREATE TABLE last_ranking
                        (record_id INTEGER,
                        ranking INT,
                        classifier TEXT,
                        query_strategy TEXT,
                        balance_strategy TEXT,
                        feature_extraction TEXT,
                        training_set INTEGER,
                        time INTEGER)''')
        cur.executemany(
            """INSERT INTO last_ranking VALUES
                                    (?, ?, ?, ?, ?, ?, ?, ?)""", last_ranking)
        con.commit()


def convert_json_last_probabilities(sql_fp, json_fp):
    """Get the last ranking from a json state and save it as the table
    'last_probabilities' in the .sql file at the location of sql_fp.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    json_fp: str/path
        Path to the json state file.
    """
    with open_state_legacy(json_fp) as json_state:
        # Get the last predicted probabilities from the state file.
        # Also get the number of record labeled and the classifier.
        last_probabilities = json_state.pred_proba

        # Put them in the format for input in the sqlite database.
        last_probabilities = [(proba, ) for proba in last_probabilities]

        with sqlite3.connect(sql_fp) as con:
            cur = con.cursor()
            cur.execute("""CREATE TABLE last_probabilities
                            (proba REAL)""")
            cur.executemany(
                """INSERT INTO last_probabilities VALUES
                                        (?)""", last_probabilities)

            con.commit()


def get_json_state_data_hash(json_state):
    """Get the data hash from a json state."""
    return list(json_state._state_dict['data_properties'].keys())[0]


def get_json_record_table(json_state):
    """Get the record table from a json state."""
    data_hash = get_json_state_data_hash(json_state)
    record_table = json_state._state_dict['data_properties'][data_hash][
        'record_table']
    return record_table


def convert_json_feature_matrix(fp, json_fp, feature_extraction_method):
    """Get the feature matrix from a json state file. Save it in the feature
    matrices folder. Format is .npz if the matrix is sparse and .npy if the
    matrix is dense.

    Arguments
    ---------
    fp: str/path
        Project folder.
    json_fp: str/path
        Path to the json state file.
    feature_extraction_method: str
        Name of the feature extraction method.

    Returns
    -------
    pathlib.Path
        Path where the feature matrix is saved.
    """
    feature_matrices_fp = Path(fp, 'feature_matrices')
    feature_matrices_fp.mkdir()

    with open_state_legacy(json_fp) as json_state:
        data_hash = get_json_state_data_hash(json_state)
        feature_matrix = decode_feature_matrix(json_state, data_hash)
        if issparse(feature_matrix):
            save_fp = Path(feature_matrices_fp,
                           f'{feature_extraction_method}_feature_matrix.npz')
            save_npz(save_fp, feature_matrix)
        else:
            save_fp = Path(feature_matrices_fp,
                           f'{feature_extraction_method}_feature_matrix.npy')
            np.save(save_fp, feature_matrix)

    return save_fp


def convert_json_record_table(sql_fp, json_fp):
    """Get the record table and save as the table 'record_table'
    in the .sql file at results_fp.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    json_fp: str/path
        Path to the json state file.
    """

    with open_state_legacy(json_fp) as json_state:
        record_table = get_json_record_table(json_state)

    # Convert record_table to list of tuples.
    record_table = [(record_id, ) for record_id in record_table]

    con = sqlite3.connect(sql_fp)
    cur = con.cursor()
    cur.execute('''CREATE TABLE record_table
                    (record_id INT)''')
    cur.executemany(
        """INSERT INTO record_table VALUES
                                (?)""", record_table)
    con.commit()
    con.close()


def convert_json_results_to_sql(sql_fp, json_fp):
    """Convert the result of a json state file to a sqlite database."""
    with open_state_legacy(json_fp, read_only=True) as sf:
        with sqlite3.connect(sql_fp) as con:
            cur = con.cursor()

            # Create the results table.
            cur.execute('''CREATE TABLE results
                            (record_id INTEGER,
                            label INTEGER,
                            classifier TEXT,
                            query_strategy TEXT,
                            balance_strategy TEXT,
                            feature_extraction TEXT,
                            training_set INTEGER,
                            labeling_time INTEGER,
                            notes TEXT)''')
            # TODO(State): models_training?

            # Index (row number) of record being labeled.
            sf_indices = [
                int(sample_data[0])
                for query in range(len(sf._state_dict['results']))
                for sample_data in sf._state_dict['results'][query]['labelled']
            ]

            # Record ids of the labeled records.
            record_table = get_json_record_table(sf)
            sf_record_ids = [int(record_table[idx]) for idx in sf_indices]

            # Label of record.
            sf_labels = [
                int(sample_data[1])
                for query in range(len(sf._state_dict['results']))
                for sample_data in sf._state_dict['results'][query]['labelled']
            ]

            # query strategy.
            sf_query_strategy = [
                sample_data[2]
                for query in range(len(sf._state_dict['results']))
                for sample_data in sf._state_dict['results'][query]['labelled']
            ]

            n_priors = sf_query_strategy.count('prior')
            n_records_labeled = len(sf_indices)
            n_non_prior_records = n_records_labeled - n_priors

            # classifier.
            classifier = sf.settings.to_dict()['model']
            sf_classifiers = [None] * n_priors + [
                f'{classifier}' for _ in range(n_non_prior_records)
            ]

            # training set.
            sf_training_sets = [-1] * n_priors + list(
                range(n_priors, n_records_labeled))

            # feature extraction.
            feature_extraction = sf.settings.to_dict()['feature_extraction']
            sf_feature_extraction = [None] * n_priors + [
                f'{feature_extraction}' for _ in range(n_non_prior_records)
            ]

            # balance strategy.
            balance_strategy = sf.settings.to_dict()['balance_strategy']
            sf_balance_strategy = [None] * n_priors + [
                f'{balance_strategy}' for _ in range(n_non_prior_records)
            ]

            # Labeling time.
            sf_time = [0 for _ in range(n_records_labeled)]

            # No notes were saved before.
            sf_notes = [None for _ in range(n_records_labeled)]

            # Check that all datasets have the same number of entries.
            lengths = [
                len(sf_record_ids),
                len(sf_labels),
                len(sf_classifiers),
                len(sf_training_sets),
                len(sf_query_strategy),
                len(sf_time),
                len(sf_feature_extraction),
                len(sf_balance_strategy),
                len(sf_notes)
            ]
            if not all([length == n_records_labeled for length in lengths]):
                raise StateError(
                    "All datasets should have the same number of entries.")

            # Create the database rows.
            db_rows = [(sf_record_ids[i], sf_labels[i], sf_classifiers[i],
                        sf_query_strategy[i], sf_balance_strategy[i],
                        sf_feature_extraction[i], sf_training_sets[i],
                        sf_time[i], sf_notes[i])
                       for i in range(n_records_labeled)]
            cur.executemany(
                """INSERT INTO results VALUES
                            (?, ?, ?, ?, ?, ?, ?, ?, ?)""", db_rows)
            con.commit()


def create_decision_changes_table(sql_fp):
    """Create an emtpy table that will contain the record_ids and new labels
    of the records whose label was changed after the original labeling action.
    Also contains the time at which the label was changed."""
    with sqlite3.connect(sql_fp) as con:
        cur = con.cursor()

        cur.execute('''CREATE TABLE decision_changes
                                    (record_id INTEGER,
                                    new_label INTEGER,
                                    time INTEGER)''')

        con.commit()



import os
from pathlib import Path


def asreview_path():
    """Get the location where projects are stored.

    Overwrite this location by specifying the ASREVIEW_PATH enviroment
    variable.
    """

    if os.environ.get("ASREVIEW_PATH", None):
        asreview_path = Path(os.environ["ASREVIEW_PATH"])
    else:
        asreview_path = Path("~", ".asreview").expanduser()

    asreview_path.mkdir(parents=True, exist_ok=True)

    return asreview_path


def list_asreview_project_paths():
    """List the projects in the asreview path"""

    file_list = []
    for x in asreview_path().iterdir():
        if x.is_dir():
            if Path(x, "project.json").exists():
                file_list.append(x)
    return file_list


def get_project_path(project_id):
    """Get the project directory.

    Arguments
    ---------
    project_id: str
        The id of the current project.
    """

    return Path(asreview_path(), project_id)
